import io
from itertools import product
import os
import re
import tempfile
import unittest
import unittest.mock

import attr

from fuzzysearch import find_near_matches_in_file
from fuzzysearch.common import Match

from tests.compat import b
import tests.test_find_near_matches


class TestSearchFile(unittest.TestCase):
    def test_file_openers(self):
        import codecs
        import io

        needle = 'PATTERN'
        haystack = '---PATERN---'

        with tempfile.NamedTemporaryFile(mode='wb', delete=False) as f:
            filename = f.name
            f.write(b(haystack))
        self.addCleanup(os.remove, filename)

        def test_file_bytes(f):
            self.assertEqual(find_near_matches_in_file(b(needle), f, max_l_dist=1),
                             [Match(3, 9, 1, b('PATERN'))])

        def test_file_unicode(f):
            self.assertEqual(find_near_matches_in_file(needle, f, max_l_dist=1),
                             [Match(3, 9, 1, 'PATERN')])

        with open(filename, 'rb') as f:
            test_file_bytes(f)

        with open(filename, 'r') as f:
            test_file_unicode(f)

        with codecs.open(filename, 'rb') as f:
            test_file_bytes(f)

        with codecs.open(filename, 'r') as f:
            test_file_unicode(f)

        with io.open(filename, 'rb') as f:
            test_file_bytes(f)

        with io.open(filename, 'r') as f:
            test_file_unicode(f)

    def test_unicode_encodings(self):
        needle = 'PATTERN'
        haystack = '---PATERN---'

        for encoding in ['ascii', 'latin-1', 'latin1', 'utf-8', 'utf-16']:
            with self.subTest(encoding=encoding):
                with tempfile.NamedTemporaryFile(mode='wb', delete=False) as f:
                    filename = f.name
                    f.write(haystack.encode(encoding))
                self.addCleanup(os.remove, filename)
                with io.open(filename, 'r', encoding=encoding) as f:
                    self.assertEqual(
                        find_near_matches_in_file(needle, f, max_l_dist=1),
                        [Match(3, 9, 1, 'PATERN')],
                    )

    def test_subsequence_split_between_chunks(self):
        with tempfile.NamedTemporaryFile(mode='wb', delete=False) as f:
            filename = f.name
        self.addCleanup(os.remove, filename)

        for needle, haystack_match, max_l_dist, expected_matches in [
            (b('PATTERN'), b('PATERN'), 0, []),
            (b('PATTERN'), b('PATERN'), 1, [Match(0, 6, 1, b('PATERN'))]),
            (b('PATTERN'), b('PATERN'), 2, [Match(0, 6, 1, b('PATERN'))]),
            (b('PATTERN'), b('PATTERN'), 0, [Match(0, 7, 0, b('PATERN'))]),
        ]:
            for chunk_size, delta in product(
                    [100, 2**10, 2**12, 2**18, 2**20],
                    sorted({-len(needle), -len(needle) + 1, -4, -2, -1, 0, 1})
            ):
                if len(needle) // (max_l_dist + 1) < 3:
                    # no ngrams search, so skip long searches which will be slow
                    if chunk_size > 2**10:
                        continue
                with self.subTest(
                        needle=needle, haystack_match=haystack_match, max_l_dist=max_l_dist,
                        chunk_size=chunk_size, delta=delta,
                ):
                    haystack = bytearray(chunk_size + 100)
                    haystack[chunk_size + delta:chunk_size + delta + len(haystack_match)] = haystack_match
                    with open(filename, 'wb') as f:
                        f.write(haystack)

                    with open(filename, 'rb') as f:
                        self.assertEqual(
                            find_near_matches_in_file(needle, f, max_l_dist=max_l_dist, _chunk_size=chunk_size),
                            [attr.evolve(match,
                                         start=match.start + chunk_size + delta,
                                         end=match.end + chunk_size + delta,
                                         matched=haystack_match)
                             for match in expected_matches]
                        )

                        f.seek(0)

                        self.assertEqual(
                            find_near_matches_in_file(needle, f, max_l_dist=max_l_dist, _chunk_size=chunk_size // 2),
                            [attr.evolve(match,
                                         start=match.start + chunk_size + delta,
                                         end=match.end + chunk_size + delta,
                                         matched=haystack_match)
                             for match in expected_matches]
                        )

                    with open(filename, 'r') as f:
                        _needle = needle.decode('utf-8')
                        self.assertEqual(
                            find_near_matches_in_file(_needle, f, max_l_dist=max_l_dist, _chunk_size=chunk_size),
                            [attr.evolve(match,
                                         start=match.start + chunk_size + delta,
                                         end=match.end + chunk_size + delta,
                                         matched=haystack.decode('utf-8'))
                             for match in expected_matches]
                        )

                    with io.open(filename, 'r', encoding='ascii') as f:
                        self.assertEqual(
                            find_near_matches_in_file(needle.decode('ascii'), f, max_l_dist=max_l_dist, _chunk_size=chunk_size),
                            [attr.evolve(match,
                                         start=match.start + chunk_size + delta,
                                         end=match.end + chunk_size + delta,
                                         matched=haystack.decode('utf-8'))
                             for match in expected_matches]
                        )

                        f.seek(0)

                        self.assertEqual(
                            find_near_matches_in_file(needle.decode('ascii'), f, max_l_dist=max_l_dist, _chunk_size=chunk_size // 2),
                            [attr.evolve(match,
                                         start=match.start + chunk_size + delta,
                                         end=match.end + chunk_size + delta,
                                         matched=haystack.decode('utf-8'))
                             for match in expected_matches]
                        )


# WARNING, DARK MAGIC AHEAD!
#
# Dynamically generate sub-classes of the TestFindNearMatchesAs* classes
# in tests.test_find_near_matches.  These sub-classes's setUp() monkey-patches
# fuzzysearch.find_near_matches(), replacing it with a wrapper for
# fuzzysearch.find_near_matches_in_file() using temporary files.
#
# TODO: replace this with something less magical

for class_name in filter(
    lambda x: re.match(r'^TestFindNearMatchesAs[a-zA-Z0-9_]+$', x),
    dir(tests.test_find_near_matches)
):
    new_class_name = class_name.replace('FindNearMatches',
                                        'FindNearMatchesInFile')
    orig_class = getattr(tests.test_find_near_matches, class_name)

    def setUp(self, _orig_class=orig_class):
        _orig_class.setUp(self)

        def find_near_matches_dropin(subsequence, sequence, *args, **kwargs):
            if isinstance(sequence, (tuple, list)):
                self.skipTest('skipping word-list tests with find_near_matches_in_file')
            try:
                from Bio.Seq import Seq
            except ImportError:
                pass
            else:
                if isinstance(sequence, Seq):
                    self.skipTest('skipping BioPython Seq tests with find_near_matches_in_file')

            tempfilepath = tempfile.mktemp()
            if isinstance(sequence, str):
                f = io.open(tempfilepath, 'w+', encoding='utf-8')
            else:
                f = open(tempfilepath, 'w+b')
            try:
                f.write(sequence)
                f.seek(0)
                return find_near_matches_in_file(subsequence, f, *args, **kwargs)
            finally:
                f.close()
                os.remove(tempfilepath)

        patcher = unittest.mock.patch(
            'tests.test_find_near_matches.find_near_matches',
            find_near_matches_dropin)
        self.addCleanup(patcher.stop)
        patcher.start()

    new_class = type(new_class_name, (orig_class,), {'setUp': setUp})
    globals()[new_class_name] = new_class



import unittest
import unittest.mock

from tests.test_search_exact import TestSearchExactBase
from tests.test_substitutions_only import TestSubstitionsOnlyBase
from tests.test_levenshtein import TestFindNearMatchesLevenshteinBase

from fuzzysearch import find_near_matches, Match
from fuzzysearch.common import FuzzySearchBase


class MockSearchClassFailsUnlessDefined(FuzzySearchBase):
    UNDEFINED = object()

    def __init__(self):
        self.return_value = self.UNDEFINED
        self.call_count = 0
        self.call_args = None

    def search(self, *args, **kwargs):
        self.call_count += 1
        self.call_args = (args, kwargs)

        if self.return_value is self.UNDEFINED:
            raise Exception('Undefined mock function called!')
        else:
            return self.return_value


class TestFindNearMatches(unittest.TestCase):
    def patch_concrete_search_classes(self):
        self.mock_search_exact = MockSearchClassFailsUnlessDefined()
        self.mock_find_near_matches_levenshtein = \
            MockSearchClassFailsUnlessDefined()
        self.mock_find_near_matches_substitutions = \
            MockSearchClassFailsUnlessDefined()
        self.mock_find_near_matches_generic = \
            MockSearchClassFailsUnlessDefined()

        patcher = unittest.mock.patch.multiple(
            'fuzzysearch',
            ExactSearch=self.mock_search_exact,
            LevenshteinSearch=
                self.mock_find_near_matches_levenshtein,
            SubstitutionsOnlySearch=
                self.mock_find_near_matches_substitutions,
            GenericSearch=
                self.mock_find_near_matches_generic,
        )
        self.addCleanup(patcher.stop)
        patcher.start()

    def test_no_limitations(self):
        with self.assertRaises(Exception):
            find_near_matches('a', 'a')

    def test_unlimited_parameter(self):
        with self.assertRaises(Exception):
            find_near_matches('a', 'a', max_substitutions=1)

        with self.assertRaises(Exception):
            find_near_matches('a', 'a', max_insertions=1)

        with self.assertRaises(Exception):
            find_near_matches('a', 'a', max_deletions=1)

        with self.assertRaises(Exception):
            find_near_matches('a', 'a', max_substitutions=1, max_insertions=1)

        with self.assertRaises(Exception):
            find_near_matches('a', 'a', max_substitutions=1, max_deletions=1)

        with self.assertRaises(Exception):
            find_near_matches('a', 'a', max_insertions=1, max_deletions=1)

    def test_all_zero(self):
        self.patch_concrete_search_classes()
        self.mock_search_exact.return_value = [Match(42, 43, 0, 'x')]
        self.assertEqual(
            find_near_matches('a', 'a', 0, 0, 0, 0),
            [Match(42, 43, 0, 'x')],
        )
        self.assertEqual(self.mock_search_exact.call_count, 1)

    def test_zero_max_l_dist(self):
        self.patch_concrete_search_classes()
        self.mock_search_exact.return_value = [Match(42, 43, 0, 'x')]

        call_count = 0
        for (max_subs, max_ins, max_dels) in [
            (1, 0, 0),
            (0, 1, 0),
            (0, 0, 1),
            (1, 1, 0),
            (1, 0, 1),
            (0, 1, 1),
            (1, 1, 1),
        ]:
            with self.subTest('max_l_dist=0, max_subs={0}, max_ins={1}, max_dels={2}'.format(
                    max_subs, max_ins, max_dels)):
                self.assertEqual(
                    find_near_matches('a', 'a', max_subs, max_ins, max_dels, 0),
                    [Match(42, 43, 0, 'x')],
                )
                call_count += 1
                self.assertEqual(self.mock_search_exact.call_count, call_count)

    def test_all_zero_except_max_l_dist(self):
        self.patch_concrete_search_classes()
        self.mock_search_exact.return_value = [Match(42, 43, 0, 'x')]

        self.assertEqual(
            find_near_matches('a', 'a', 0, 0, 0, 1),
            [Match(42, 43, 0, 'x')],
        )
        self.assertEqual(self.mock_search_exact.call_count, 1)

    def test_all_none_except_max_l_dist(self):
        self.patch_concrete_search_classes()
        self.mock_find_near_matches_levenshtein.return_value = \
            [Match(42, 43, 0, 'x')]

        self.assertEqual(
            find_near_matches('a', 'a', max_l_dist=1),
            [Match(42, 43, 0, 'x')],
        )
        self.assertEqual(self.mock_find_near_matches_levenshtein.call_count, 1)

    def test_levenshtein(self):
        """test cases where 0 < max_l_dist <= max(others)"""
        # in these cases, find_near_matches should call
        # find_near_matches_levenshtein
        self.patch_concrete_search_classes()
        self.mock_find_near_matches_levenshtein.return_value = \
            [unittest.mock.sentinel.SENTINEL]

        self.assertEqual(
            find_near_matches('a', 'a', 1, 1, 1, 1),
            [unittest.mock.sentinel.SENTINEL],
        )
        self.assertEqual(self.mock_find_near_matches_levenshtein.call_count, 1)

        self.assertEqual(
            find_near_matches('a', 'a', 2, 2, 2, 2),
            [unittest.mock.sentinel.SENTINEL],
        )
        self.assertEqual(self.mock_find_near_matches_levenshtein.call_count, 2)

        self.assertEqual(
            find_near_matches('a', 'a', 5, 3, 7, 2),
            [unittest.mock.sentinel.SENTINEL],
        )
        self.assertEqual(self.mock_find_near_matches_levenshtein.call_count, 3)

    def test_only_substitutions(self):
        self.patch_concrete_search_classes()
        self.mock_find_near_matches_substitutions.return_value = \
            [Match(42, 43, 0, 'x')]

        self.assertEqual(
            find_near_matches('a', 'a', 1, 0, 0),
            [Match(42, 43, 0, 'x')],
        )
        self.assertEqual(
            self.mock_find_near_matches_substitutions.call_count,
            1,
        )

        self.assertEqual(
            find_near_matches('a', 'a', 1, 0, 0, 1),
            [Match(42, 43, 0, 'x')],
        )
        self.assertEqual(
            self.mock_find_near_matches_substitutions.call_count,
            2,
        )

    def test_generic(self):
        self.patch_concrete_search_classes()
        self.mock_find_near_matches_generic.return_value = \
            [Match(42, 43, 0, 'x')]

        self.assertEqual(
            find_near_matches('a', 'a', 1, 1, 1),
            [Match(42, 43, 0, 'x')],
        )
        self.assertEqual(
            self.mock_find_near_matches_generic.call_count,
            1,
        )

        self.assertEqual(
            find_near_matches('a', 'a', 1, 1, 1, 2),
            [Match(42, 43, 0, 'x')],
        )
        self.assertEqual(
            self.mock_find_near_matches_generic.call_count,
            2,
        )


class TestFindNearMatchesAsLevenshtein(TestFindNearMatchesLevenshteinBase,
                                       unittest.TestCase):
    def search(self, subsequence, sequence, max_l_dist):
        return find_near_matches(subsequence, sequence, max_l_dist=max_l_dist)


class TestFindNearMatchesAsSearchExact(TestSearchExactBase,
                                       unittest.TestCase):
    def search(self, subsequence, sequence, start_index=0, end_index=None):
        if end_index is None:
            end_index = len(sequence)
        sequence = sequence[start_index:end_index]
        return [
            start_index + match.start
            for match in find_near_matches(subsequence, sequence, max_l_dist=0)
        ]

    @classmethod
    def get_supported_sequence_types(cls):
        from tests.test_search_exact import TestSearchExact
        return TestSearchExact.get_supported_sequence_types()


class TestFindNearMatchesAsSubstitutionsOnly(TestSubstitionsOnlyBase,
                                             unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        return find_near_matches(subsequence, sequence,
                                 max_insertions=0, max_deletions=0,
                                 max_substitutions=max_subs)

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        return self.assertEqual(search_results, expected_outcomes, *args, **kwargs)


from tests.test_generic_search import TestGenericSearch
class TestFindNearMatchesAsGeneric(TestGenericSearch,
                                   unittest.TestCase):
    def search(self, pattern, sequence, max_subs, max_ins, max_dels,
               max_l_dist=None):
        return find_near_matches(pattern, sequence,
                                 max_subs, max_ins, max_dels, max_l_dist)
del TestGenericSearch



import re
import unittest

from fuzzysearch.common import Match, consolidate_overlapping_matches
from fuzzysearch.levenshtein import find_near_matches_levenshtein, \
    find_near_matches_levenshtein_linear_programming as fnm_levenshtein_lp
from fuzzysearch.levenshtein_ngram import \
    _expand, _py_expand_short, _expand_long, \
    find_near_matches_levenshtein_ngrams as fnm_levenshtein_ngrams


def longstr(string):
    return re.sub(r'\s+', '', string)


class TestFuzzySearch(unittest.TestCase):
    def test_empty_sequence(self):
        self.assertEqual(
            list(fnm_levenshtein_lp('PATTERN', '', max_l_dist=0)),
            [],
        )

    def test_empty_subsequence_exeption(self):
        with self.assertRaises(ValueError):
            list(fnm_levenshtein_lp('', 'TEXT', max_l_dist=0))

    def test_match_identical_sequence(self):
        matches = \
            list(fnm_levenshtein_lp('PATTERN', 'PATTERN', max_l_dist=0))
        self.assertEqual(matches, [Match(start=0, end=len('PATTERN'), dist=0,
                                         matched='PATTERN')])

    def test_double_first_item(self):
        sequence = 'abcddefg'
        pattern = 'def'
        matches = \
            list(fnm_levenshtein_lp(pattern, sequence, max_l_dist=1))
        self.assertIn(Match(start=4, end=7, dist=0, matched=pattern), matches)

    def test_missing_second_item(self):
        sequence = 'abcdefg'
        pattern = 'bde'
        matches = \
            list(fnm_levenshtein_lp(pattern, sequence, max_l_dist=1))
        self.assertIn(Match(start=1, end=5, dist=1, matched='bcde'), matches)

    def test_dna_search(self):
        # see: http://stackoverflow.com/questions/19725127/
        text = ''.join('''\
            GACTAGCACTGTAGGGATAACAATTTCACACAGGTGGACAATTACATTGAAAATCACAGATTGGT
            CACACACACATTGGACATACATAGAAACACACACACATACATTAGATACGAACATAGAAACACAC
            ATTAGACGCGTACATAGACACAAACACATTGACAGGCAGTTCAGATGATGACGCCCGACTGATAC
            TCGCGTAGTCGTGGGAGGCAAGGCACACAGGGGATAGG
            '''.split())
        pattern = 'TGCACTGTAGGGATAACAAT'

        matches = list(fnm_levenshtein_lp(pattern, text, max_l_dist=2))

        self.assertTrue(len(matches) > 0)
        self.assertIn(Match(start=3, end=24, dist=1, matched=text[3:24]),
                      matches)


class TestExpandBase(object):
    expand = None  # override in sub-classes!

    def test_both_empty(self):
        self.assertEqual(self.expand('', '', 0), (0, 0))

    def test_empty_subsequence(self):
        self.assertEqual(self.expand('', 'TEXT', 0), (0, 0))

    def test_empty_sequence(self):
        self.assertEqual(self.expand('PATTERN', '', 0), (None, None))
        self.assertEqual(self.expand('PATTERN', '', 6), (None, None))
        self.assertEqual(self.expand('PATTERN', '', 7), (7, 0))
        self.assertEqual(self.expand('PATTERN', '', 8), (7, 0))

    def test_identical(self):
        self.assertEqual(self.expand('abc', 'abc', 0), (0, 3))
        self.assertEqual(self.expand('abc', 'abc', 1), (0, 3))
        self.assertEqual(self.expand('abc', 'abc', 2), (0, 3))

    def test_first_item_missing(self):
        self.assertEqual(self.expand('abcd', 'bcd', 0), (None, None))
        self.assertEqual(self.expand('abcd', 'bcd', 1), (1, 3))
        self.assertEqual(self.expand('abcd', 'bcd', 2), (1, 3))

    def test_second_item_missing(self):
        self.assertEqual(self.expand('abcd', 'acd', 0), (None, None))
        self.assertEqual(self.expand('abcd', 'acd', 1), (1, 3))
        self.assertEqual(self.expand('abcd', 'acd', 2), (1, 3))

    def test_second_before_last_item_missing(self):
        self.assertEqual(self.expand('abcd', 'abd', 0), (None, None))
        self.assertEqual(self.expand('abcd', 'abd', 1), (1, 3))
        self.assertEqual(self.expand('abcd', 'abd', 2), (1, 3))

    def test_last_item_missing(self):
        self.assertEqual(self.expand('abcd', 'abc', 0), (None, None))
        self.assertEqual(self.expand('abcd', 'abc', 1), (1, 3))
        self.assertEqual(self.expand('abcd', 'abc', 2), (1, 3))

    def test_completely_different(self):
        self.assertEqual(self.expand('abc', 'def', 0), (None, None))

    def test_startswith(self):
        self.assertEqual(self.expand('abc', 'abcd', 0), (0, 3))
        self.assertEqual(self.expand('abc', 'abcd', 1), (0, 3))
        self.assertEqual(self.expand('abc', 'abcd', 2), (0, 3))

    def test_missing_at_start_middle_and_end(self):
        self.assertEqual(self.expand('abcd', '-ab-cd-', 0), (None, None))
        self.assertEqual(self.expand('abcd', '-ab-cd-', 1), (None, None))
        self.assertEqual(self.expand('abcd', '-ab-cd-', 2), (2, 6))
        self.assertEqual(self.expand('abcd', '-ab-cd-', 3), (2, 6))

    def test_no_common_chars(self):
        self.assertEqual(self.expand('abc', 'de', 2), (None, None))
        self.assertEqual(self.expand('abc', 'de', 3)[0], 3)
        self.assertEqual(self.expand('abc', 'de', 4)[0], 3)

    def test_long_needle(self):
        self.assertEqual(
            self.expand('abcdefghijklmnop', 'abcdefg-hijk-mnopqrst', 0),
            (None, None),
        )
        self.assertEqual(
            self.expand('abcdefghijklmnop', 'abcdefg-hijk-mnopqrst', 1),
            (None, None),
        )
        self.assertEqual(
            self.expand('abcdefghijklmnop', 'abcdefg-hijk-mnopqrst', 2),
            (2, 17),
        )
        self.assertEqual(
            self.expand('abcdefghijklmnop', 'abcdefg-hijk-mnopqrst', 3),
            (2, 17),
        )

        self.assertEqual(
            self.expand('abcdefghijklmnop', 'abcdefg-hijk-mnop', 3),
            (2, 17),
        )

        self.assertEqual(
            self.expand('abcdefghijklmnop', '-bcdefg-hijk-mnop', 3),
            (3, 17),
        )
        self.assertEqual(
            self.expand('abcdefghijklmnop', '-abcdefg-hijk-mnop', 3),
            (3, 18),
        )

        self.assertEqual(
            self.expand('abcdefghijklmnop', 'abc---defg-hijk-mnopqrst', 8),
            (5, 20),
        )


class TestExpand(TestExpandBase, unittest.TestCase):
    expand = staticmethod(_expand)


class TestPyExpandShort(TestExpandBase, unittest.TestCase):
    expand = staticmethod(_py_expand_short)


try:
    from fuzzysearch._levenshtein_ngrams import c_expand_short
except ImportError:
    pass
else:
    class TestCExpandShort(TestExpandBase, unittest.TestCase):
        expand = staticmethod(c_expand_short)


class TestExpandLong(TestExpandBase, unittest.TestCase):
    expand = staticmethod(_expand_long)


class TestFindNearMatchesLevenshteinBase(object):
    def search(self, subsequence, sequence, max_l_dist):
        raise NotImplementedError
    
    test_cases_data = {
        # name: (needle, haystack, [
        #   (max_l_dist, [(start, end, dist), ...]),
        # ])
        'identical sequence': ('PATTERN', 'PATTERN', [
            (0, [(0, 7, 0)]),
        ]),
        'substring': ('PATTERN', '----------PATTERN---------', [
            (0, [(10, 17, 0)]),
            (1, [(10, 17, 0)]),
            (2, [(10, 17, 0)]),
        ]),
        'double first item': ('def', 'abcddefg', [
            (1, [(4, 7, 0)]),
        ]),
        'double last item': ('def', 'abcdeffg', [
            (1, [(3, 6, 0)]),
        ]),
        'double first items': ('defgh', 'abcdedefghi', [
            (3, [(5, 10, 0)]),
        ]),
        'double last items': ('cdefgh', 'abcdefghghi', [
            (3, [(2, 8, 0)]),
        ]),
        'missing second item': ('bde', 'abcdefg', [
            (1, [(1, 5, 1)]),
        ]),
        'missing second to last item': ('bce', 'abcdefg', [
            (1, [(1, 5, 1)]),
            (2, [(1, 5, 1)]),
        ]),
        'one missing in middle': ('PATTERN', '----------PATERN---------', [
            (0, []),
            (1, [(10, 16, 1)]),
            (2, [(10, 16, 1)]),
        ]),
        'one changed in middle': ('PATTERN', '----------PAT-ERN---------', [
            (0, []),
            (1, [(10, 17, 1)]),
            (2, [(10, 17, 1)]),
        ]),
        'one extra in middle': ('PATTERN', '----------PATT-ERN---------', [
            (0, []),
            (1, [(10, 18, 1)]),
            (2, [(10, 18, 1)]),
        ]),
        'one extra repeating in middle': ('PATTERN', '----------PATTTERN---------', [
            (0, []),
            (1, [(10, 18, 1)]),
            (2, [(10, 18, 1)]),
        ]),
        'one extra repeating at end': ('PATTERN', '----------PATTERNN---------', [
            (0, [(10, 17, 0)]),
            (1, [(10, 17, 0)]),
            (2, [(10, 17, 0)]),
        ]),
        'one missing at end': ('defg', 'abcdef', [
            (1, [(3, 6, 1)]),
        ]),
        'highly repetetive': ('a' * 9, 'a' * 7 + 'xx', [
            (1, []),
            (2, [(0, 9, 2)]),
        ]),
        'DNA search': (
            'TGCACTGTAGGGATAACAAT',
            longstr('''
                GACTAGCACTGTAGGGATAACAATTTCACACAGGTGGACAATTACATTGAAAATCACAGATTG
                GTCACACACACATTGGACATACATAGAAACACACACACATACATTAGATACGAACATAGAAAC
                ACACATTAGACGCGTACATAGACACAAACACATTGACAGGCAGTTCAGATGATGACGCCCGAC
                TGATACTCGCGTAGTCGTGGGAGGCAAGGCACACAGGGGATAGG
            '''),
            [
                (2, [(3, 24, 1)]),
            ]
        ),
        # see:
        # * BioPython archives from March 14th, 2014
        #   http://lists.open-bio.org/pipermail/biopython/2014-March/009030.html
        # * https://github.com/taleinat/fuzzysearch/issues/3
        'protein search 1': (
            'GGGTTLTTSS',
            longstr('''
                XXXXXXXXXXXXXXXXXXXGGGTTVTTSSAAAAAAAAAAAAAGGGTTLTTSSAAAAAAAAAAA
                AAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBGGGTTLTTSS
            '''),
            [
                (0, [(42, 52, 0), (99, 109, 0)]),
                (1, [(19, 29, 1), (42, 52, 0), (99, 109, 0)]),
                (2, [(19, 29, 1), (42, 52, 0), (99, 109, 0)]),
            ]
        ),
        'protein search 2': (
            'GGGTTLTTSS',
            longstr('''
                XXXXXXXXXXXXXXXXXXXGGGTTVTTSSAAAAAAAAAAAAAGGGTTVTTSSAAAAAAAAAAA
                AAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBGGGTTLTTSS
            '''),
            [
                (0, [(99, 109, 0)]),
                (1, [(19, 29, 1), (42, 52, 1), (99, 109, 0)]),
                (2, [(19, 29, 1), (42, 52, 1), (99, 109, 0)]),
            ]
        ),
        'list of words': (
            "over a lazy dog".split(),
            "the big brown fox jumped over the lazy dog".split(),
            [
                (0, []),
                (1, [(5, 9, 1)]),
                (2, [(5, 9, 1)]),
            ]
        ),
    }

    def test_cases(self):
        for name, data in self.test_cases_data.items():
            substring, text, max_l_dist2expected_matches = data
            with self.subTest(name=name):
                for max_l_dist, expected_matches in max_l_dist2expected_matches:
                    self.assertEqual(
                        self.search(substring, text, max_l_dist=max_l_dist),
                        [Match(*x, matched=text[x[0]:x[1]])
                         for x in expected_matches],
                    )

    def test_empty_sequence(self):
        self.assertEqual(self.search('PATTERN', '', max_l_dist=0), [])

    def test_empty_subsequence_exeption(self):
        with self.assertRaises(ValueError):
            self.search('', 'TEXT', max_l_dist=0)

    def test_all_different(self):
        for max_l_dist in [0, 1, 2, 3]:
            self.assertEqual(
                self.search('AAAA', 'ZZZZ', max_l_dist),
                [],
            )

        matches = self.search('AAAA', 'ZZZZ', max_l_dist=4)
        self.assertGreater(len(matches), 0)
        self.assertTrue(all(match.dist == 4 for match in matches))


class TestFindNearMatchesLevenshteinNgrams(TestFindNearMatchesLevenshteinBase,
                                           unittest.TestCase):
    def search(self, subsequence, sequence, max_l_dist):
        if max_l_dist >= len(subsequence):
            self.skipTest(
                'skipping ngram search with max_l_dist >= len(subsequence)')
        return consolidate_overlapping_matches(
            fnm_levenshtein_ngrams(subsequence, sequence, max_l_dist)
        )


class TestFindNearMatchesLevenshteinLP(TestFindNearMatchesLevenshteinBase,
                                       unittest.TestCase):
    def search(self, subsequence, sequence, max_l_dist):
        return consolidate_overlapping_matches(
            fnm_levenshtein_lp(subsequence, sequence, max_l_dist)
        )


class TestFindNearMatchesLevenshtein(TestFindNearMatchesLevenshteinBase,
                                     unittest.TestCase):
    def search(self, subsequence, sequence, max_l_dist):
        return consolidate_overlapping_matches(
            find_near_matches_levenshtein(subsequence, sequence, max_l_dist)
        )



import unittest

from fuzzysearch.common import group_matches, Match, get_best_match_in_group, \
    count_differences_with_maximum, consolidate_overlapping_matches
from fuzzysearch.substitutions_only import \
    has_near_match_substitutions as hnm_subs, \
    find_near_matches_substitutions as fnm_subs, \
    find_near_matches_substitutions_lp as fnm_subs_lp, \
    has_near_match_substitutions_lp as hnm_subs_lp, \
    find_near_matches_substitutions_ngrams as fnm_subs_ngrams, \
    has_near_match_substitutions_ngrams as hnm_subs_ngrams

from tests.compat import b
from tests.utils import skip_if_arguments_arent_byteslike


class TestSubstitionsOnlyBase(object):
    def search(self, subsequence, sequence, max_subs):
        raise NotImplementedError

    def expectedOutcomes(self, search_result, expected_outcomes, *args, **kwargs):
        raise NotImplementedError

    def test_empty_sequence(self):
        self.expectedOutcomes(self.search(b('PATTERN'), b(''), max_subs=0), [])

    def test_empty_subsequence_exeption(self):
        with self.assertRaises(ValueError):
            self.search(b(''), b('TEXT'), max_subs=0)

    def test_match_identical_sequence(self):
        self.expectedOutcomes(
            self.search(b('PATTERN'), b('PATTERN'), max_subs=0),
            [Match(start=0, end=len('PATTERN'), dist=0, matched=b('PATTERN'))],
        )

    def test_substring(self):
        substring = b('PATTERN')
        text = b('aaaaaaaaaaPATTERNaaaaaaaaa')
        expected_match = Match(start=10, end=17, dist=0, matched=b('PATTERN'))

        self.expectedOutcomes(
            self.search(substring, text, max_subs=0),
            [expected_match],
        )
        self.expectedOutcomes(
            self.search(substring, text, max_subs=1),
            [expected_match],
        )
        self.expectedOutcomes(
            self.search(substring, text, max_subs=2),
            [expected_match],
        )

    def test_double_first_item(self):
        self.expectedOutcomes(
            self.search(b('def'), b('abcddefg'), max_subs=1),
            [Match(start=4, end=7, dist=0, matched=b('def'))],
        )

        self.expectedOutcomes(
            self.search(b('def'), b('abcddefg'), max_subs=2),
            [Match(start=3, end=6, dist=2, matched=b('dde')),
             Match(start=4, end=7, dist=0, matched=b('def'))],
        )

    def test_two_identical(self):
        self.expectedOutcomes(
            self.search(b('abc'), b('abcabc'), max_subs=1),
            [Match(start=0, end=3, dist=0, matched=b('abc')),
             Match(start=3, end=6, dist=0, matched=b('abc'))],
        )

        self.expectedOutcomes(
            self.search(b('abc'), b('abcXabc'), max_subs=1),
            [Match(start=0, end=3, dist=0, matched=b('abc')),
             Match(start=4, end=7, dist=0, matched=b('abc'))],
        )

    def test_one_changed_in_middle(self):
        substring = b('abcdefg')
        pattern = b('abcXefg')
        expected_match = Match(start=0, end=7, dist=1, matched=pattern)

        self.expectedOutcomes(
            self.search(substring, pattern, max_subs=0),
            [],
        )

        self.expectedOutcomes(
            self.search(substring, pattern, max_subs=1),
            [expected_match],
        )

        self.expectedOutcomes(
            self.search(substring, pattern, max_subs=2),
            [expected_match],
        )

    def test_one_missing_in_middle(self):
        substring = b('PATTERN')
        text = b('aaaaaaaaaaPATERNaaaaaaaaa')

        for max_subs in [0, 1, 2]:
            self.expectedOutcomes(
                self.search(substring, text, max_subs=max_subs),
                [],
            )

    def test_one_changed_in_middle2(self):
        substring = b('PATTERN')
        text = b('aaaaaaaaaaPATtERNaaaaaaaaa')
        expected_match = Match(start=10, end=17, dist=1, matched=b('PATtERN'))

        self.expectedOutcomes(
            self.search(substring, text, max_subs=0),
            [],
        )
        self.expectedOutcomes(
            self.search(substring, text, max_subs=1),
            [expected_match],
        )
        self.expectedOutcomes(
            self.search(substring, text, max_subs=2),
            [expected_match],
        )

    def test_one_extra_in_middle(self):
        substring = b('PATTERN')
        text = b('aaaaaaaaaaPATTXERNaaaaaaaaa')

        for max_subs in [0, 1, 2]:
            self.expectedOutcomes(
                self.search(substring, text, max_subs=max_subs),
                [],
            )

    def test_all_different(self):
        substring = b('AAAA')
        text = b('ZZZZ')

        for max_subs in [0, 1, 2, 3]:
            self.expectedOutcomes(
                self.search(substring, text, max_subs=max_subs),
                [],
            )

        for max_subs in [4, 5]:
            self.expectedOutcomes(
                self.search(substring, text, max_subs=max_subs),
                [Match(start=0, end=4, dist=4, matched=b('ZZZZ'))],
            )

    def test_dna_search(self):
        # see: http://stackoverflow.com/questions/19725127/
        text = b(''.join('''\
            GACTAGCACTGTAGGGATAACAATTTCACACAGGTGGACAATTACATTGAAAATCACAGATTGGT
            CACACACACATTGGACATACATAGAAACACACACACATACATTAGATACGAACATAGAAACACAC
            ATTAGACGCGTACATAGACACAAACACATTGACAGGCAGTTCAGATGATGACGCCCGACTGATAC
            TCGCGTAGTCGTGGGAGGCAAGGCACACAGGGGATAGG
            '''.split()))
        pattern = b('TGCACTGTAGGGATAACAAT')

        self.expectedOutcomes(
            self.search(pattern, text, max_subs=2),
            [Match(start=4, end=24, dist=1, matched=text[4:24])],
        )

    def test_protein_search1(self):
        # see:
        # * BioPython archives from March 14th, 2014
        #   http://lists.open-bio.org/pipermail/biopython/2014-March/009030.html
        # * https://github.com/taleinat/fuzzysearch/issues/3
        text = b(''.join('''\
            XXXXXXXXXXXXXXXXXXXGGGTTVTTSSAAAAAAAAAAAAAGGGTTLTTSSAAAAAAAAAAAA
            AAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBGGGTTLTTSS
        '''.split()))
        pattern = b("GGGTTLTTSS")

        self.expectedOutcomes(
            self.search(pattern, text, max_subs=0),
            [Match(start=42, end=52, dist=0, matched=text[42:52]),
             Match(start=99, end=109, dist=0, matched=text[99:109])],
        )

        self.expectedOutcomes(
            self.search(pattern, text, max_subs=1),
            [Match(start=19, end=29, dist=1, matched=text[19:29]),
             Match(start=42, end=52, dist=0, matched=text[42:52]),
             Match(start=99, end=109, dist=0, matched=text[99:109])],
        )

        self.expectedOutcomes(
            self.search(pattern, text, max_subs=2),
            [Match(start=19, end=29, dist=1, matched=text[19:29]),
             Match(start=42, end=52, dist=0, matched=text[42:52]),
             Match(start=99, end=109, dist=0, matched=text[99:109])],
        )

    def test_protein_search2(self):
        # see:
        # * BioPython archives from March 14th, 2014
        #   http://lists.open-bio.org/pipermail/biopython/2014-March/009030.html
        # * https://github.com/taleinat/fuzzysearch/issues/3
        text = b(''.join('''\
            XXXXXXXXXXXXXXXXXXXGGGTTVTTSSAAAAAAAAAAAAAGGGTTVTTSSAAAAAAAAAAA
            AAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBGGGTTLTTSS
        '''.split()))
        pattern = b("GGGTTLTTSS")

        self.expectedOutcomes(
            self.search(pattern, text, max_subs=0),
            [Match(start=99, end=109, dist=0, matched=text[99:109])],
        )

        self.expectedOutcomes(
            self.search(pattern, text, max_subs=1),
            [Match(start=19, end=29, dist=1, matched=text[19:29]),
             Match(start=42, end=52, dist=1, matched=text[42:52]),
             Match(start=99, end=109, dist=0, matched=text[99:109])],
        )

        self.expectedOutcomes(
            self.search(pattern, text, max_subs=2),
            [Match(start=19, end=29, dist=1, matched=text[19:29]),
             Match(start=42, end=52, dist=1, matched=text[42:52]),
             Match(start=99, end=109, dist=0, matched=text[99:109])],
        )

    def test_missing_at_beginning(self):
        self.expectedOutcomes(
            self.search(b("ATTEST"), b("TESTOSTERONE"), max_subs=2),
            [],
        )

    def test_unicode_substring(self):
        pattern = '\u03A3\u0393'
        text = '\u03A0\u03A3\u0393\u0394'
        self.expectedOutcomes(
            self.search(pattern, text, max_subs=0),
            [Match(1, 3, 0, matched=text[1:3])]
        )

    def test_max_substitutions_gte_subseq_len(self):
        for max_subs in [1, 2, 5]:
            self.expectedOutcomes(
                self.search(b('b'), b('abc'), max_subs),
                [Match(0, 1, 1, b('a')),
                 Match(1, 2, 0, b('b')),
                 Match(2, 3, 1, b('c'))]
            )
        for extra_subs in [0, 1, 7]:
            self.expectedOutcomes(
                self.search(b('PATTERN'), b('PATTERN'), len('PATTERN') + extra_subs),
                [Match(0, len('PATTERN'), 0, b('PATTERN'))]
            )


class TestFindNearMatchesSubstitions(TestSubstitionsOnlyBase,
                                     unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        return fnm_subs(subsequence, sequence, max_subs)

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        best_from_grouped_results = [
            get_best_match_in_group(group)
            for group in group_matches(search_results)
        ]
        best_from_grouped_exepected_outcomes = [
            get_best_match_in_group(group)
            for group in group_matches(expected_outcomes)
        ]
        return self.assertEqual(best_from_grouped_results,
                                best_from_grouped_exepected_outcomes,
                                *args, **kwargs)


class TestFindNearMatchesSubstitionsLinearProgramming(TestSubstitionsOnlyBase,
                                                      unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        return list(fnm_subs_lp(subsequence, sequence, max_subs))

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        return self.assertEqual(search_results, expected_outcomes, *args, **kwargs)


class TestFindNearMatchesSubstitionsNgrams(TestSubstitionsOnlyBase,
                                           unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        if max_subs >= len(subsequence):
            self.skipTest("avoiding calling fnm_subs_ngrams() " +
                          "with max_subs >= len(subsequence)")
        return fnm_subs_ngrams(subsequence, sequence, max_subs)

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        return self.assertEqual(
            consolidate_overlapping_matches(search_results),
            consolidate_overlapping_matches(expected_outcomes),
            *args, **kwargs)


class TestHasNearMatchSubstitionsOnlyBase(TestSubstitionsOnlyBase):
    def search(self, subsequence, sequence, max_subs):
        raise NotImplementedError

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        return self.assertEqual(bool(search_results),
                                bool(expected_outcomes),
                                *args, **kwargs)


class TestHasNearMatchSubstitionsOnly(TestHasNearMatchSubstitionsOnlyBase,
                                      unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        return hnm_subs(subsequence, sequence, max_subs)


class TestHasNearMatchSubstitionsOnlyNgrams(TestHasNearMatchSubstitionsOnlyBase,
                                            unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        if max_subs >= len(subsequence):
            self.skipTest("avoiding calling hnm_subs_ngrams() " +
                          "with max_subs >= len(subsequence)")
        return hnm_subs_ngrams(subsequence, sequence, max_subs)


class TestHasNearMatchSubstitionsOnlyLp(TestHasNearMatchSubstitionsOnlyBase,
                                        unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        return hnm_subs_lp(subsequence, sequence, max_subs)


try:
    from fuzzysearch._substitutions_only import \
        substitutions_only_has_near_matches_lp_byteslike as \
            hnm_subs_lp_byteslike, \
        substitutions_only_find_near_matches_lp_byteslike as \
            fnm_subs_lp_byteslike, \
        substitutions_only_has_near_matches_ngrams_byteslike as \
            hnm_subs_ngrams_byteslike, \
        substitutions_only_find_near_matches_ngrams_byteslike as \
            fnm_subs_ngrams_byteslike
except ImportError:
    pass
else:
    class TestHasNearMatchesSubstitionsLpByteslike(
            TestHasNearMatchSubstitionsOnlyBase,
            unittest.TestCase
    ):
        @skip_if_arguments_arent_byteslike
        def search(self, subsequence, sequence, max_subs):
            return hnm_subs_lp_byteslike(subsequence, sequence,
                                         max_subs)

    class TestHasNearMatchesSubstitionsNgramsByteslike(
            TestHasNearMatchSubstitionsOnlyBase,
            unittest.TestCase
    ):
        @skip_if_arguments_arent_byteslike
        def search(self, subsequence, sequence, max_subs):
            if max_subs >= len(subsequence):
                self.skipTest("avoiding calling hnm_subs_ngrams_byteslike() " +
                              "with max_subs >= len(subsequence)")
            return hnm_subs_ngrams_byteslike(subsequence, sequence,
                                             max_subs)

    class TestFindNearMatchesSubstitionsLpByteslike(
            TestSubstitionsOnlyBase,
            unittest.TestCase
    ):
        @skip_if_arguments_arent_byteslike
        def search(self, subsequence, sequence, max_subs):
            results = fnm_subs_lp_byteslike(subsequence, sequence,
                                            max_subs)
            matches = [
                Match(
                    index,
                    index + len(subsequence),
                    count_differences_with_maximum(
                        sequence[index:index+len(subsequence)],
                        subsequence,
                        max_subs + 1,
                    ),
                    matched=sequence[index:index+len(subsequence)]
                )
                for index in results
            ]
            return matches

        def expectedOutcomes(self, search_results, expected_outcomes,
                             *args, **kwargs):
            return self.assertEqual(search_results, expected_outcomes,
                                    *args, **kwargs)

    class TestFindNearMatchesSubstitionsNgramsByteslike(
            TestSubstitionsOnlyBase,
            unittest.TestCase
    ):
        @skip_if_arguments_arent_byteslike
        def search(self, subsequence, sequence, max_subs):
            results = fnm_subs_ngrams_byteslike(subsequence, sequence,
                                                max_subs)
            matches = [
                Match(
                    index,
                    index + len(subsequence),
                    count_differences_with_maximum(
                        sequence[index:index+len(subsequence)],
                        subsequence,
                        max_subs + 1,
                    ),
                    matched=sequence[index:index+len(subsequence)]
                )
                for index in results
            ]
            return [
                get_best_match_in_group(group)
                for group in group_matches(matches)
            ]

        def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
            return self.assertEqual(
                consolidate_overlapping_matches(search_results),
                consolidate_overlapping_matches(expected_outcomes),
                *args, **kwargs)



import unittest

from fuzzysearch.search_exact import search_exact
from tests.compat import b


class TestSearchExactBase(object):
    def search(self, subsequence, sequence, start_index=0, end_index=None):
        raise NotImplementedError

    def test_empty_sequence(self):
        self.assertEqual(self.search('PATTERN', ''), [])

    def test_empty_subsequence(self):
        with self.assertRaises(ValueError):
            self.search('', 'TEXT')

    def test_match_identical_sequence(self):
        self.assertEqual(self.search('PATTERN', 'PATTERN'), [0])

    def test_substring(self):
        substring = 'PATTERN'
        text = 'aaaaaaaaaaPATTERNaaaaaaaaa'
        self.assertEqual(self.search(substring, text), [10])

    def test_double_first_item(self):
        self.assertEqual(self.search('def', 'abcddefg'), [4])

    def test_missing_second_item(self):
        self.assertEqual(self.search('bde', 'abcdefg'), [])

    def test_completely_different(self):
        self.assertEqual(self.search('abc', 'def'), [])

    def test_startswith(self):
        self.assertEqual(self.search('abc', 'abcd'), [0])

    def test_endswith(self):
        self.assertEqual(self.search('bcd', 'abcd'), [1])

    @classmethod
    def get_supported_sequence_types(cls):
        raise NotImplementedError

    def test_identical(self):
        # search for a pattern in itself, should match once at index 0
        for initializer in self.get_supported_sequence_types():
            pattern = initializer('abc')
            with self.subTest("search_exact({0!r}, {0!r})".format(pattern)):
                self.assertEqual(self.search(pattern, pattern), [0])

    def test_subsequence(self):
        # search for a pattern appearing once at index 4
        for initializer in self.get_supported_sequence_types():
            pattern = initializer('abc')
            sequence = initializer('-ab-abc-ab-')
            with self.subTest("search_exact({0!r}, {1!r})".format(pattern, sequence)):
                self.assertEqual(self.search(pattern, sequence), [4])

    def test_multiple_matches(self):
        # search for a pattern appearing at indexes 1, 5 and 9
        for initializer in self.get_supported_sequence_types():
            pattern = initializer('abc')
            sequence = initializer('-abc-abc-abc-')
            with self.subTest("search_exact({0!r}, {1!r})".format(pattern, sequence)):
                self.assertEqual(self.search(pattern, sequence), [1, 5, 9])

    def test_outside_range_limits(self):
        for initializer in self.get_supported_sequence_types():
            pattern = initializer('abc')
            sequence = initializer('-abc-abc-abc')
            with self.subTest("search_exact({0!r}, {1!r}, {2}, {3})".format(pattern, sequence, 0, 3)):
                self.assertEqual(self.search(pattern, sequence, 0, 3), [])
            with self.subTest("search_exact({0!r}, {1!r}, {2}, {3})".format(pattern, sequence, 0, 4)):
                self.assertEqual(self.search(pattern, sequence, 0, 4), [1])
            with self.subTest("search_exact({0!r}, {1!r}, {2}, {3})".format(pattern, sequence, 0, 7)):
                self.assertEqual(self.search(pattern, sequence, 0, 7), [1])
            with self.subTest("search_exact({0!r}, {1!r}, {2}, {3})".format(pattern, sequence, 0, 8)):
                self.assertEqual(self.search(pattern, sequence, 0, 8), [1, 5])
            with self.subTest("search_exact({0!r}, {1!r}, {2}, {3})".format(pattern, sequence, 0, 11)):
                self.assertEqual(self.search(pattern, sequence, 0, 11), [1, 5])
            with self.subTest("search_exact({0!r}, {1!r}, {2}, {3})".format(pattern, sequence, 0, 12)):
                self.assertEqual(self.search(pattern, sequence, 0, 12), [1, 5, 9])

            with self.subTest("search_exact({0!r}, {1!r}, {2})".format(pattern, sequence, 1)):
                self.assertEqual(self.search(pattern, sequence, 1), [1, 5, 9])
            with self.subTest("search_exact({0!r}, {1!r}, {2})".format(pattern, sequence, 2)):
                self.assertEqual(self.search(pattern, sequence, 2), [5, 9])
            with self.subTest("search_exact({0!r}, {1!r}, {2})".format(pattern, sequence, 5)):
                self.assertEqual(self.search(pattern, sequence, 5), [5, 9])
            with self.subTest("search_exact({0!r}, {1!r}, {2})".format(pattern, sequence, 6)):
                self.assertEqual(self.search(pattern, sequence, 6), [9])

            with self.subTest("search_exact({0!r}, {1!r}, {2}, {3})".format(pattern, sequence, 4, 10)):
                self.assertEqual(self.search(pattern, sequence, 4, 10), [5])

            with self.subTest("search_exact({0!r}, {1!r}, {2}, {3})".format(pattern, sequence, 3, 7)):
                self.assertEqual(self.search(pattern, sequence, 3, 7), [])


class TestSearchExact(TestSearchExactBase, unittest.TestCase):
    def search(self, subsequence, sequence, start_index=0, end_index=None):
        return list(search_exact(subsequence, sequence, start_index, end_index))

    @classmethod
    def get_supported_sequence_types(cls):
        types_to_test = [b, str, list, tuple]

        try:
            from Bio.Seq import Seq
            from Bio.Alphabet import IUPAC
        except ImportError:
            pass
        else:
            types_to_test.append(Seq)
            types_to_test.append(
                lambda text: Seq(text.replace('b', 'g').replace('-', 't'),
                                 alphabet=IUPAC.unambiguous_dna))

        return types_to_test

    def test_unicode_subsequence(self):
        self.assertEqual(self.search('\u03A3\u0393', '\u03A0\u03A3\u0393\u0394'), [1])


try:
    from fuzzysearch._common import search_exact_byteslike
except ImportError:
    pass
else:
    class TestSearchExactByteslike(TestSearchExactBase, unittest.TestCase):
        def search(self, subsequence, sequence, start_index=0, end_index=None):
            if isinstance(subsequence, str):
                try:
                    subsequence = subsequence.encode('ascii')
                except UnicodeEncodeError:
                    raise self.skipTest("skipping test with non-ascii-encodable string for byteslike function")
            if isinstance(sequence, str):
                try:
                    sequence = sequence.encode('ascii')
                except UnicodeEncodeError:
                    raise self.skipTest("skipping test with non-ascii-encodable string for byteslike function")

            if end_index is not None:
                return search_exact_byteslike(subsequence, sequence, start_index, end_index)
            else:
                return search_exact_byteslike(subsequence, sequence, start_index)

        @classmethod
        def get_supported_sequence_types(cls):
            types_to_test = [b]
            return types_to_test

        def test_input_argument_handling(self):
            self.assertEqual(search_exact_byteslike(b'abc', b'abc'), [0])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', 0), [0])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', 1), [])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', 0, 3), [0])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', 0, end_index=3), [0])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', end_index=3, start_index=0), [0])
            self.assertEqual(search_exact_byteslike(subsequence=b'abc', sequence=b'abc',
                                                    start_index=0, end_index=3), [0])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', 0, 4), [0])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', 0, -1), [0])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', 0, 2), [])
            self.assertEqual(search_exact_byteslike(b'abc', b'abc', 2, 1), [])

            with self.assertRaises(Exception):
                search_exact_byteslike(b'abc', subsequence=b'abc')

            with self.assertRaises(Exception):
                search_exact_byteslike(b'abc', b'abc', 0, start_index=0)



import unittest

from tests.compat import b
from tests.test_levenshtein import TestFindNearMatchesLevenshteinBase
from fuzzysearch.common import Match, get_best_match_in_group, group_matches, LevenshteinSearchParams, consolidate_overlapping_matches
from tests.test_substitutions_only import TestSubstitionsOnlyBase, \
    TestHasNearMatchSubstitionsOnlyBase
from fuzzysearch.generic_search import \
    _find_near_matches_generic_linear_programming as fnm_generic_lp, \
    find_near_matches_generic_ngrams as fnm_generic_ngrams, \
    has_near_match_generic_ngrams as hnm_generic_ngrams, \
    find_near_matches_generic


class TestGenericSearchLpAsLevenshtein(TestFindNearMatchesLevenshteinBase,
                                       unittest.TestCase):
    def search(self, subsequence, sequence, max_l_dist):
        search_params = LevenshteinSearchParams(max_l_dist, max_l_dist,
                                                max_l_dist, max_l_dist)
        return consolidate_overlapping_matches(
            fnm_generic_lp(subsequence, sequence, search_params)
        )


class TestGenericSearchNgramsAsLevenshtein(TestFindNearMatchesLevenshteinBase,
                                           unittest.TestCase):
    def search(self, subsequence, sequence, max_l_dist):
        if max_l_dist >= len(subsequence):
            self.skipTest("avoiding calling fnm_generic_ngrams() " +
                          "with max_l_dist >= len(subsequence)")
        search_params = LevenshteinSearchParams(max_l_dist, max_l_dist,
                                                max_l_dist, max_l_dist)
        return consolidate_overlapping_matches(
            fnm_generic_ngrams(subsequence, sequence, search_params)
        )


class TestGenericSearchLpAsSubstitutionsOnly(TestSubstitionsOnlyBase,
                                             unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        search_params = LevenshteinSearchParams(max_subs, 0, 0, max_subs)
        return list(
            fnm_generic_lp(subsequence, sequence, search_params)
        )

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        return self.assertEqual(search_results, expected_outcomes, *args, **kwargs)


class TestGenericSearchNgramsAsSubstitutionsOnly(TestSubstitionsOnlyBase,
                                                 unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        if max_subs >= len(subsequence):
            self.skipTest("avoiding calling fnm_generic_ngrams() " +
                          "with max_subs >= len(subsequence)")
        search_params = LevenshteinSearchParams(max_subs, 0, 0, max_subs)
        return consolidate_overlapping_matches(
            fnm_generic_ngrams(subsequence, sequence, search_params)
        )

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        best_from_grouped_results = [
            get_best_match_in_group(group)
            for group in group_matches(search_results)
        ]
        best_from_grouped_exepected_outcomes = [
            get_best_match_in_group(group)
            for group in group_matches(expected_outcomes)
        ]
        return self.assertEqual(best_from_grouped_results,
                                best_from_grouped_exepected_outcomes,
                                *args, **kwargs)


class TestGenericSearchBase(object):
    def search(self, pattern, sequence, max_subs, max_ins, max_dels,
               max_l_dist=None):
        raise NotImplementedError

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        raise NotImplementedError

    def test_empty_sequence(self):
        self.assertEqual(self.search(b('PATTERN'), b(''), 0, 0, 0, 0), [])

    def test_empty_subsequence_exeption(self):
        with self.assertRaises(ValueError):
            self.search(b(''), b('TEXT'), 0, 0, 0, 0)

    def test_match_identical_sequence(self):
        self.assertEqual(
            self.search(b('PATTERN'), b('PATTERN'), 0, 0, 0, 0),
            [Match(start=0, end=7, dist=0, matched=b('PATTERN'))],
        )

    def test_short_substring(self):
        substring = b('XY')
        text = b('abcdefXYghij')
        expected_match = Match(start=6, end=8, dist=0, matched=substring)

        self.assertEqual(
            self.search(substring, text, 0, 0, 0, 0),
            [expected_match],
        )

    def test_substring(self):
        substring = b('PATTERN')
        text = b('aaaaaaaaaaPATTERNaaaaaaaaa')
        expected_match = Match(start=10, end=17, dist=0, matched=substring)

        self.assertEqual(
            self.search(substring, text, 0, 0, 0, 0),
            [expected_match],
        )

    def test_substring_in_long_text(self):
        substring = b('PATTERN')
        text = b(''.join([x.strip() for x in '''\
            FySijRLMtLLWkMnWxTbzIWuxOUbfAahWYKUlOZyhoQhfExJPOSwXxBLrlqdoUwpRW
            FEtHFiepnOTbkttuagADQaUTvkvKzvqaFaMnAPfolPpmXitKLDQhAqDOJwFzdcKmk
            cfVStxZGDUbrHjrDwVVRihbklyfqLJjrzGuhVGDzgSpCHXvaGPHebbcUAnAgfqqpA
            uMOowtptcoQUeAbdqJAmieLDxCrOPivbSwmriQwfFCDTXbswFqClZPnSkDkCyvPCi
            bmAjVGnuVsrZlPypglXlVVQKzMpQuWQynOLGDqwrAnsvYTcArkEhFpEgahWVQGOvv
            CTvbYZRVqqPCDRsyWeTVgANxZIyVAtENnndbsHzpEcPUfqCBUroIGRNEIMHYIZANy
            LeeVKEwihbvWZVOWPeAlmNKnhhoEPIcpDJDzPOYHSltxhSsZeeWMqtAnuSoFOIrqB
            EPUFIlKkpamljHylnTIWqaESoWbYESVPEeZtlAzpInuwFaNIYUvzpJNIlPtuOjUuT
            efaGnOXvQeHdaRPrdHCepPATTERNDdnkzuLHQcVWKpgHhGifBySAkWkthrzfZDHDU
            HJxjpLXseKuldLRftyctGvVKyrRTUCRAakjwTSWivGdksOZabnkBoRtMstlNwXcwg
            UCFLaWFxjqjasOfNeThrbubVGtyYRROYUOTMUmeSdJcBKxVXiaWDZoHyKtQRXwpVO
            pEmlpdzKWkFpDtHHdImhDJIXwxzjwyNLaTgPLHmcyhJGqncCblxALMdPEDaRtGFMg
            BskUxPGATTLKMFeIjgFJpudyMWlASyFSiaDWrOCgRfwjfpMYfuNQIqzvZbguWsnaq
            tRaXcxavobetBbbfMDjstQLjoJLwiajVRKhFVspIdgrmTMEBbjtpMnSpTkmFcRBZZ
            GUOWnesGgZeKkIQhlxlRPTtjUbbpaPlmxeiBdUKHHApgvEybUwWwXCoXFsauNiINm
            AGATFdcaHzgoRpbBFhKdJkLMF'''.splitlines()]))
        expected_match = Match(start=541, end=548, dist=0, matched=substring)

        self.assertEqual(
            self.search(substring, text, 0, 0, 0, 0),
            [expected_match],
        )

    def test_single_substitution_in_long_text(self):
        substring = b('PATTERN')
        text = b(''.join([x.strip() for x in '''\
            FySijRLMtLLWkMnWxTbzIWuxOUbfAahWYKUlOZyhoQhfExJPOSwXxBLrlqdoUwpRW
            FEtHFiepnOTbkttuagADQaUTvkvKzvqaFaMnAPfolPpmXitKLDQhAqDOJwFzdcKmk
            cfVStxZGDUbrHjrDwVVRihbklyfqLJjrzGuhVGDzgSpCHXvaGPHebbcUAnAgfqqpA
            uMOowtptcoQUeAbdqJAmieLDxCrOPivbSwmriQwfFCDTXbswFqClZPnSkDkCyvPCi
            bmAjVGnuVsrZlPypglXlVVQKzMpQuWQynOLGDqwrAnsvYTcArkEhFpEgahWVQGOvv
            CTvbYZRVqqPCDRsyWeTVgANxZIyVAtENnndbsHzpEcPUfqCBUroIGRNEIMHYIZANy
            LeeVKEwihbvWZVOWPeAlmNKnhhoEPIcpDJDzPOYHSltxhSsZeeWMqtAnuSoFOIrqB
            EPUFIlKkpamljHylnTIWqaESoWbYESVPEeZtlAzpInuwFaNIYUvzpJNIlPtuOjUuT
            efaGnOXvQeHdaRPrdHCepPATXERNDdnkzuLHQcVWKpgHhGifBySAkWkthrzfZDHDU
            HJxjpLXseKuldLRftyctGvVKyrRTUCRAakjwTSWivGdksOZabnkBoRtMstlNwXcwg
            UCFLaWFxjqjasOfNeThrbubVGtyYRROYUOTMUmeSdJcBKxVXiaWDZoHyKtQRXwpVO
            pEmlpdzKWkFpDtHHdImhDJIXwxzjwyNLaTgPLHmcyhJGqncCblxALMdPEDaRtGFMg
            BskUxPGATTLKMFeIjgFJpudyMWlASyFSiaDWrOCgRfwjfpMYfuNQIqzvZbguWsnaq
            tRaXcxavobetBbbfMDjstQLjoJLwiajVRKhFVspIdgrmTMEBbjtpMnSpTkmFcRBZZ
            GUOWnesGgZeKkIQhlxlRPTtjUbbpaPlmxeiBdUKHHApgvEybUwWwXCoXFsauNiINm
            AGATFdcaHzgoRpbBFhKdJkLMF'''.splitlines()]))
        expected_match = Match(start=541, end=548, dist=1,
                               matched=text[541:548])

        self.assertEqual(
            self.search(substring, text, 1, 0, 0, 1),
            [expected_match],
        )

        self.assertEqual(
            self.search(substring, text, 1, 1, 1, 1),
            [expected_match],
        )

    def test_double_first_item(self):
        self.expectedOutcomes(
            self.search(b('def'), b('abcddefg'), 0, 0, 0, 0),
            [Match(start=4, end=7, dist=0, matched=b('def'))],
        )

        self.expectedOutcomes(
            self.search(b('def'), b('abcddefg'), 1, 0, 0, 1),
            [Match(start=4, end=7, dist=0, matched=b('def'))],
        )

        self.expectedOutcomes(
            self.search(b('def'), b('abcddefg'), 0, 0, 1, 1),
            [Match(start=4, end=7, dist=0, matched=b('def')),
             Match(start=5, end=7, dist=1, matched=b('ef'))]
        )

        self.expectedOutcomes(
            self.search(b('def'), b('abcddefg'), 0, 1, 0, 1),
            [Match(start=3, end=7, dist=1, matched=b('ddef')),
             Match(start=4, end=7, dist=0, matched=b('def'))],
        )

    def test_missing_second_item(self):
        self.assertEqual(
            self.search(b('bde'), b('abcdefg'), 0, 1, 0, 1),
            [Match(start=1, end=5, dist=1, matched=b('bcde'))],
        )

        self.assertEqual(
            self.search(b('bde'), b('abcdefg'), 0, 0, 0, 0),
            [],
        )

        self.assertEqual(
            self.search(b('bde'), b('abcdefg'), 1, 0, 0, 1),
            [Match(start=2, end=5, dist=1, matched=b('cde'))],
        )

        self.assertEqual(
            self.search(b('bde'), b('abcdefg'), 0, 0, 1, 1),
            [Match(start=3, end=5, dist=1, matched=b('de'))],
        )

    def test_null_bytes(self):
        self.assertEqual(
            self.search(b('abc'), b('xx\0abcxx'), 0, 0, 0, 0),
            [Match(start=3, end=6, dist=0, matched=b('abc'))],
        )

        self.assertEqual(
            self.search(b('a\0b'), b('xxa\0bcxx'), 0, 0, 0, 0),
            [Match(start=2, end=5, dist=0, matched=b('a\0b'))],
        )

    def test_valid_none_arguments_with_defined_max_l_dist(self):
        # expect no exception when max_l_dist is not None and some or all other
        # values are None
        N = None
        for (max_subs, max_ins, max_dels) in [
            (N, 0, 0),
            (0, N, 0),
            (0, 0, N),
            (0, N, N),
            (N, 0, N),
            (N, N, 0),
            (N, N, N),
        ]:
            with self.subTest('max_subs={0}, max_ins={1}, max_dels={2}, max_l_dist=0'.format(
                    max_subs, max_ins, max_dels)):
                self.assertEqual(
                    self.search(b('a'), b('b'), max_subs, max_ins, max_dels, 0),
                    [],
                )

    def test_only_max_l_dist_none(self):
        # expect no exception when only max_l_dist is None
        self.assertEqual(
            self.search(b('a'), b('b'), 0, 0, 0, None),
            [],
        )

    def test_no_deletion(self):
        self.assertEqual(
            self.search(b('PATTERN'), b('---PATERN---'), None, None, 0, 1),
            []
        )

    def test_only_deletions(self):
        self.expectedOutcomes(
            self.search(b('TESTabc'), b('TEST123'), 0, 0, 5, None),
            [
                Match(start=0, end=4, dist=3, matched=b('TEST')),
                Match(start=1, end=4, dist=4, matched=b('EST')),
                Match(start=2, end=4, dist=5, matched=b('ST')),
            ],
        )

    def test_invalid_none_arguments(self):
        # check that an exception is raised when max_l_dist is None as well as
        # at least one other limitation
        N = None
        for (max_subs, max_ins, max_dels) in [
            (N, 0, 0),
            (0, N, 0),
            (0, 0, N),
            (0, N, N),
            (N, 0, N),
            (N, N, 0),
            (N, N, N),
        ]:
            with self.subTest('max_subs={0}, max_ins={1}, max_dels={2}, max_l_dist=None'.format(
                    max_subs, max_ins, max_dels)):
                with self.assertRaises(ValueError):
                    self.search(b('a'), b('b'), max_subs, max_ins, max_dels, None)


class TestGenericSearch(TestGenericSearchBase, unittest.TestCase):
    def search(self, pattern, sequence, max_subs, max_ins, max_dels,
               max_l_dist=None):
        search_params = LevenshteinSearchParams(max_subs, max_ins,
                                                max_dels, max_l_dist)
        return consolidate_overlapping_matches(
            find_near_matches_generic(pattern, sequence, search_params)
        )

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        consolidated_expected_outcomes = \
            consolidate_overlapping_matches(expected_outcomes)
        return self.assertEqual(search_results, consolidated_expected_outcomes)

    def test_non_string_sequences(self):
        supported_types = [list, tuple]
        for klass in supported_types:
            with self.subTest(klass.__name__):
                self.expectedOutcomes(self.search(klass([1, 2, 3]), klass([1, 2, 3]), 0, 0, 0, 0),
                                      [Match(start=0, end=3, dist=0,
                                             matched=klass([1, 2, 3]))])
                self.expectedOutcomes(self.search(klass([1, 2, 3]), klass([1, 2, 3]), 1, 1, 1, 1),
                                      [Match(start=0, end=3, dist=0,
                                             matched=klass([1, 2, 3]))])
                self.expectedOutcomes(self.search(klass([1, 2, 3]), klass([1, 2, 4]), 0, 0, 0, 0),
                                      [])
                self.expectedOutcomes(self.search(klass([1, 2, 3]), klass([1, 2, 4]), 1, 1, 1, 1),
                                      [Match(start=0, end=3, dist=1,
                                             matched=klass([1, 2, 4]))])
                self.expectedOutcomes(self.search(klass([1, 2, 3]), klass([1, 2, 4]), 0, 0, 1, 1),
                                      [Match(start=0, end=2, dist=1,
                                             matched=klass([1, 2]))])

    def test_list_of_words_one_missing(self):
        subsequence = "jumped over the a lazy dog".split()
        sequence = "the big brown fox jumped over the lazy dog".split()
        for params, expected_outcomes in [
            ((0, 0, 0, 0), []),
            ((1, 0, 0, 1), []),
            ((0, 1, 0, 1), []),
            ((0, 0, 1, 1), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
            ((1, 1, 1, 1), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
            ((2, 2, 2, 2), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
        ]:
            self.expectedOutcomes(
                self.search(subsequence, sequence, *params),
                expected_outcomes,
            )

    def test_list_of_words_one_extra(self):
        subsequence = "jumped over lazy dog".split()
        sequence = "the big brown fox jumped over the lazy dog".split()
        for params, expected_outcomes in [
            ((0, 0, 0, 0), []),
            ((1, 0, 0, 1), []),
            ((0, 1, 0, 1), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
            ((0, 0, 1, 1), []),
            ((1, 1, 1, 1), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
            ((2, 2, 2, 2), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
        ]:
            self.expectedOutcomes(
                self.search(subsequence, sequence, *params),
                expected_outcomes,
            )

    def test_list_of_words_one_substituted(self):
        subsequence = "jumped over my lazy dog".split()
        sequence = "the big brown fox jumped over the lazy dog".split()
        for params, expected_outcomes in [
            ((0, 0, 0, 0), []),
            ((1, 0, 0, 1), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
            ((0, 1, 0, 1), []),
            ((0, 0, 1, 1), []),
            # substitution = insertion + deletion; dist = 1 !!
            ((0, 1, 1, 1), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
            ((1, 1, 1, 1), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
            ((2, 2, 2, 2), [Match(start=4, end=9, dist=1,
                                  matched="jumped over the lazy dog".split())]),
        ]:
            self.expectedOutcomes(
                self.search(subsequence, sequence, *params),
                expected_outcomes,
            )


class TestNgramsBase(object):
    def test_subseq_length_less_than_max_l_dist(self):
        with self.assertRaises(ValueError):
            self.search(b('b'), b('abc'), 2, 2, 2, 2)

        with self.assertRaises(ValueError):
            self.search(b('b'), b('abc'), 5, 5, 5, 5)

        with self.assertRaises(ValueError):
            self.search(b('PATTERN'), b('PATTERN'),
                        len('PATTERN') + 1,
                        len('PATTERN') + 1,
                        len('PATTERN') + 1,
                        len('PATTERN') + 1,
                        )

        with self.assertRaises(ValueError):
            self.search(b('PATTERN'), b('PATTERN'),
                        len('PATTERN') + 7,
                        len('PATTERN') + 7,
                        len('PATTERN') + 7,
                        len('PATTERN') + 7,
                        )


class TestGenericSearchLp(TestGenericSearchBase, unittest.TestCase):
    def search(self, pattern, sequence, max_subs, max_ins, max_dels,
               max_l_dist=None):
        search_params = LevenshteinSearchParams(max_subs, max_ins,
                                                max_dels, max_l_dist)
        return list(
            fnm_generic_lp(pattern, sequence, search_params)
        )

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        self.assertEqual(search_results, expected_outcomes, *args, **kwargs)

    def test_double_first_item_two_results(self):
        self.expectedOutcomes(
            self.search(b('def'), b('abcddefg'), 0, 1, 0, 1),
            [Match(start=3, end=7, dist=1, matched=b('ddef')),
             Match(start=4, end=7, dist=0, matched=b('def'))],
        )

    def test_missing_second_item_complex(self):
        self.expectedOutcomes(
            self.search(b('bde'), b('abcdefg'), 1, 1, 1, 1),
            [Match(start=1, end=5, dist=1, matched=b('bcde')),
             Match(start=2, end=5, dist=1, matched=b('cde')),
             Match(start=3, end=5, dist=1, matched=b('de'))],
        )

        self.assertTrue(
            {
                Match(start=1, end=5, dist=1, matched=b('bcde')),
                Match(start=2, end=5, dist=1, matched=b('cde')),
                Match(start=3, end=5, dist=1, matched=b('de')),
                Match(start=2, end=5, dist=2, matched=b('bcd')),
            }.issubset(set(
                self.search(b('bde'), b('abcdefg'), 1, 1, 1, 3),
            ))
        )


class TestGenericSearchNgrams(TestGenericSearchBase,
                              TestNgramsBase,
                              unittest.TestCase):
    def search(self, pattern, sequence, max_subs, max_ins, max_dels, max_l_dist=None):
        search_params = LevenshteinSearchParams(max_subs, max_ins,
                                                max_dels, max_l_dist)
        return consolidate_overlapping_matches(
            fnm_generic_ngrams(pattern, sequence, search_params)
        )

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        consolidated_results = [
            get_best_match_in_group(group)
            for group in group_matches(search_results)
        ]
        consolidated_expected_outcomes = \
            consolidate_overlapping_matches(expected_outcomes)
        return self.assertEqual(consolidated_results,
                                consolidated_expected_outcomes,
                                *args, **kwargs)

    def test_missing_second_item_complex(self):
        self.assertTrue(
            set(self.search(b('bde'), b('abcdefg'), 1, 1, 1, 1)).issubset([
                Match(start=1, end=5, dist=1, matched=b('bcde')),
                Match(start=2, end=5, dist=1, matched=b('cde')),
                Match(start=3, end=5, dist=1, matched=b('de')),
            ])
        )


class TestHasNearMatchGenericNgramsAsSubstitutionsOnly(
    TestHasNearMatchSubstitionsOnlyBase,
):
    def search(self, subsequence, sequence, max_subs):
        return hnm_generic_ngrams(subsequence, sequence,
                                  LevenshteinSearchParams(max_subs, 0, 0, max_subs))


class TestHasNearMatchGenericNgrams(TestGenericSearchBase,
                                    TestNgramsBase,
                                    unittest.TestCase):
    def search(self, pattern, sequence, max_subs, max_ins, max_dels,
               max_l_dist=None):
        return hnm_generic_ngrams(pattern, sequence,
                                  LevenshteinSearchParams(max_subs, max_ins, max_dels, max_l_dist))

    def expectedOutcomes(self, search_results, expected_outcomes, *args, **kwargs):
        self.assertEqual(bool(search_results),
                         bool(expected_outcomes),
                         *args, **kwargs)

    def assertEqual(self, actual_value, expected_value, *args, **kwargs):
        return super(TestHasNearMatchGenericNgrams, self).assertEqual(
            actual_value, bool(expected_value), *args, **kwargs)

    def test_missing_second_item_complex(self):
        self.assertTrue(self.search(b('bde'), b('abcdefg'), 1, 1, 1, 1))



from functools import wraps

from fuzzysearch.common import FuzzySearchBase, Match


__all__ = [
    'search_exact',
    'ExactSearch',
]


CLASSES_WITH_INDEX = (list, tuple)
CLASSES_WITH_FIND = (bytes, bytearray, str)

try:
    from Bio.Seq import Seq
except ImportError:
    pass
else:
    CLASSES_WITH_FIND += (Seq,)


def search_exact(subsequence, sequence, start_index=0, end_index=None):
    if not subsequence:
        raise ValueError('subsequence must not be empty')

    if end_index is None:
        end_index = len(sequence)

    if isinstance(sequence, CLASSES_WITH_FIND):
        def find_in_index_range(start_index):
            return sequence.find(subsequence, start_index, end_index)
    elif isinstance(sequence, CLASSES_WITH_INDEX):
        first_item = subsequence[0]
        first_item_last_index = end_index - (len(subsequence) - 1)
        def find_in_index_range(start_index):
            while True:
                try:
                    first_index = sequence.index(first_item, start_index, first_item_last_index)
                    start_index = first_index + 1
                except ValueError:
                    return -1
                for subseq_index in range(1, len(subsequence)):
                    if sequence[first_index + subseq_index] != subsequence[subseq_index]:
                        break
                else:
                    return first_index
    else:
        raise TypeError('unsupported sequence type: %s' % type(sequence))

    index = find_in_index_range(start_index)
    while index >= 0:
        yield index
        index = find_in_index_range(index + 1)


try:
    from fuzzysearch._common import search_exact_byteslike
except ImportError:
    pass
else:
    _search_exact = search_exact
    @wraps(_search_exact)
    def search_exact(subsequence, sequence, start_index=0, end_index=None):
        if end_index is None:
            end_index = len(sequence)

        try:
            return search_exact_byteslike(subsequence, sequence,
                                          start_index, end_index)
        except (TypeError, UnicodeEncodeError):
            return _search_exact(subsequence, sequence, start_index, end_index)


class ExactSearch(FuzzySearchBase):
    @classmethod
    def search(cls, subsequence, sequence, search_params):
        for index in search_exact(subsequence, sequence):
            yield Match(index, index + len(subsequence), 0,
                        sequence[index:index + len(subsequence)])

    @classmethod
    def extra_items_for_chunked_search(cls, subsequence, search_params):
        return 0



import unittest

from fuzzysearch.common import Match, group_matches, GroupOfMatches, \
    count_differences_with_maximum
from tests.compat import b


class TestGroupOfMatches(unittest.TestCase):
    def test_is_match_in_group(self):
        match = Match(2, 4, 0, 'matched')
        group = GroupOfMatches(match)
        self.assertTrue(group.is_match_in_group(match))
        self.assertTrue(group.is_match_in_group(Match(2, 4, 0, 'matched')))


class TestGroupMatches(unittest.TestCase):
    def test_separate(self):
        matches = [
            Match(start=19, end=29, dist=1, matched='x'*10),
            Match(start=42, end=52, dist=1, matched='x'*10),
            Match(start=99, end=109, dist=0, matched='x'*10),
        ]
        self.assertEqual(
            group_matches(matches),
            [{m} for m in matches],
        )

    def test_separate_with_duplicate(self):
        matches = [
            Match(start=19, end=29, dist=1, matched='x'*10),
            Match(start=42, end=52, dist=1, matched='x'*10),
            Match(start=99, end=109, dist=0, matched='x'*10),
        ]
        self.assertEqual(
            group_matches(matches + [matches[1]]),
            [{m} for m in matches],
        )


class TestCountDifferencesWithMaximumBase(object):
    def count_diffs(self, seq1, seq2, max_diffs):
        raise NotImplementedError

    def test_empty(self):
        result = self.count_diffs('', '', 1)
        self.assertEqual(result, 0)

    def test_identical_one_character(self):
        result = self.count_diffs('a', 'a', 1)
        self.assertEqual(result, 0)

    def test_identical_word(self):
        result = self.count_diffs('word', 'word', 1)
        self.assertEqual(result, 0)

    def test_identical_long(self):
        result = self.count_diffs('long'*10, 'long'*10, 1)
        self.assertEqual(result, 0)

    def test_different_less_than_max(self):
        result = self.count_diffs('abc', 'def', 4)
        self.assertEqual(result, 3)

    def test_different_more_than_max(self):
        result = self.count_diffs('abc', 'def', 2)
        self.assertEqual(result, 2)

    def test_partially_different_in_middle(self):
        result = self.count_diffs('abcdef', 'a--d-f', 4)
        self.assertEqual(result, 3)

        result = self.count_diffs('abcdef', 'a--d-f', 2)
        self.assertEqual(result, 2)

    def test_partially_different_at_start(self):
        result = self.count_diffs('abcdef', '--c-ef', 4)
        self.assertEqual(result, 3)

        result = self.count_diffs('abcdef', '--c-ef', 2)
        self.assertEqual(result, 2)

    def test_partially_different_at_end(self):
        result = self.count_diffs('abcdef', 'ab-d--', 4)
        self.assertEqual(result, 3)

        result = self.count_diffs('abcdef', 'ab-d--', 2)
        self.assertEqual(result, 2)


class TestCountDifferencesWithMaximum(TestCountDifferencesWithMaximumBase,
                                      unittest.TestCase):
    def count_diffs(self, seq1, seq2, max_diffs):
        return count_differences_with_maximum(seq1, seq2, max_diffs)


try:
    from fuzzysearch._common import count_differences_with_maximum_byteslike
except ImportError:
    pass
else:
    class TestCountDifferencesWithMaximumByteslike(
            TestCountDifferencesWithMaximumBase, unittest.TestCase):
        def count_diffs(self, seq1, seq2, max_diffs):
            return count_differences_with_maximum_byteslike(b(seq1), b(seq2),
                                                            max_diffs)



from collections import namedtuple

from fuzzysearch.common import FuzzySearchBase, Match, \
    consolidate_overlapping_matches
from fuzzysearch.levenshtein_ngram import find_near_matches_levenshtein_ngrams
from fuzzysearch.search_exact import search_exact


def find_near_matches_levenshtein(subsequence, sequence, max_l_dist):
    """Find near-matches of the subsequence in the sequence.

    This chooses a suitable fuzzy search implementation according to the given
    parameters.

    Returns a list of fuzzysearch.Match objects describing the matching parts
    of the sequence.
    """
    if not subsequence:
        raise ValueError('Given subsequence is empty!')
    if max_l_dist < 0:
        raise ValueError('Maximum Levenshtein distance must be >= 0!')

    if max_l_dist == 0:
        return [
            Match(start_index, start_index + len(subsequence), 0,
                  sequence[start_index:start_index + len(subsequence)])
            for start_index in search_exact(subsequence, sequence)
        ]

    elif len(subsequence) // (max_l_dist + 1) >= 3:
        return find_near_matches_levenshtein_ngrams(subsequence,
                                                    sequence,
                                                    max_l_dist)

    else:
        return find_near_matches_levenshtein_linear_programming(subsequence,
                                                                sequence,
                                                                max_l_dist)


Candidate = namedtuple('Candidate', ['start', 'subseq_index', 'dist'])


def make_char2first_subseq_index(subsequence, max_l_dist):
    return dict(
        (char, index)
        for (index, char) in
        reversed(list(enumerate(subsequence[:max_l_dist + 1])))
    )


def find_near_matches_levenshtein_linear_programming(subsequence, sequence,
                                                     max_l_dist):
    if not subsequence:
        raise ValueError('Given subsequence is empty!')

    subseq_len = len(subsequence)

    def make_match(start, end, dist):
        return Match(start, end, dist, matched=sequence[start:end])

    if max_l_dist >= subseq_len:
        for index in range(len(sequence) + 1):
            yield make_match(index, index, subseq_len)
        return

    # optimization: prepare some often used things in advance
    char2first_subseq_index = make_char2first_subseq_index(subsequence,
                                                           max_l_dist)

    candidates = []
    for index, char in enumerate(sequence):
        new_candidates = []

        idx_in_subseq = char2first_subseq_index.get(char, None)
        if idx_in_subseq is not None:
            if idx_in_subseq + 1 == subseq_len:
                yield make_match(index, index + 1, idx_in_subseq)
            else:
                new_candidates.append(Candidate(index, idx_in_subseq + 1, idx_in_subseq))

        for cand in candidates:
            # if this sequence char is the candidate's next expected char
            if subsequence[cand.subseq_index] == char:
                # if reached the end of the subsequence, return a match
                if cand.subseq_index + 1 == subseq_len:
                    yield make_match(cand.start, index + 1, cand.dist)
                # otherwise, update the candidate's subseq_index and keep it
                else:
                    new_candidates.append(cand._replace(
                        subseq_index=cand.subseq_index + 1,
                    ))

            # if this sequence char is *not* the candidate's next expected char
            else:
                # we can try skipping a sequence or sub-sequence char (or both),
                # unless this candidate has already skipped the maximum allowed
                # number of characters
                if cand.dist == max_l_dist:
                    continue

                # add a candidate skipping a sequence char
                new_candidates.append(cand._replace(dist=cand.dist + 1))

                if index + 1 < len(sequence) and cand.subseq_index + 1 < subseq_len:
                    # add a candidate skipping both a sequence char and a
                    # subsequence char
                    new_candidates.append(cand._replace(
                        dist=cand.dist + 1,
                        subseq_index=cand.subseq_index + 1,
                    ))

                # try skipping subsequence chars
                for n_skipped in range(1, max_l_dist - cand.dist + 1):
                    # if skipping n_skipped sub-sequence chars reaches the end
                    # of the sub-sequence, yield a match
                    if cand.subseq_index + n_skipped == subseq_len:
                        yield make_match(cand.start, index + 1,
                                         cand.dist + n_skipped)
                        break
                    # otherwise, if skipping n_skipped sub-sequence chars
                    # reaches a sub-sequence char identical to this sequence
                    # char, add a candidate skipping n_skipped sub-sequence
                    # chars
                    elif subsequence[cand.subseq_index + n_skipped] == char:
                        # if this is the last char of the sub-sequence, yield
                        # a match
                        if cand.subseq_index + n_skipped + 1 == subseq_len:
                            yield make_match(cand.start, index + 1,
                                             cand.dist + n_skipped)
                        # otherwise add a candidate skipping n_skipped
                        # subsequence chars
                        else:
                            new_candidates.append(cand._replace(
                                dist=cand.dist + n_skipped,
                                subseq_index=cand.subseq_index + 1 + n_skipped,
                            ))
                        break
                # note: if the above loop ends without a break, that means that
                # no candidate could be added / yielded by skipping sub-sequence
                # chars

        candidates = new_candidates

    for cand in candidates:
        dist = cand.dist + subseq_len - cand.subseq_index
        if dist <= max_l_dist:
            yield make_match(cand.start, len(sequence), dist)


class LevenshteinSearch(FuzzySearchBase):
    @classmethod
    def search(cls, subsequence, sequence, search_params):
        for match in find_near_matches_levenshtein(subsequence, sequence,
                                                   search_params.max_l_dist):
            yield match

    @classmethod
    def consolidate_matches(cls, matches):
        return consolidate_overlapping_matches(matches)

    @classmethod
    def extra_items_for_chunked_search(cls, subsequence, search_params):
        return search_params.max_l_dist



from functools import wraps

from attr import attrs, attrib


__all__ = [
    'Match', 'LevenshteinSearchParams',
    'count_differences_with_maximum',
    'group_matches', 'get_best_match_in_group',
    'consolidate_overlapping_matches',
]


@attrs(frozen=True, slots=True)
class Match(object):
    start = attrib(type=int, eq=True, hash=True)
    end = attrib(type=int, eq=True, hash=True)
    dist = attrib(type=int, eq=True, hash=True)
    matched = attrib(eq=False, hash=False)

    if __debug__:
        def __attrs_post_init__(self):
            if not (isinstance(self.start, int) and self.start >= 0):
                raise ValueError('start must be a non-negative integer')
            if not (isinstance(self.end, int) and self.end >= self.start):
                raise ValueError('end must be an integer no smaller than start')
            if not (isinstance(self.dist, int) and self.dist >= 0):
                print(self.dist)
                raise ValueError('dist must be a non-negative integer')
            if self.matched is None:
                raise ValueError('matched must be supplied')


@attrs(frozen=True, slots=True)
class LevenshteinSearchParams(object):
    """Parameter data-class for Levenshtein-distance fuzzy searches."""
    max_substitutions = attrib(default=None)
    max_insertions = attrib(default=None)
    max_deletions = attrib(default=None)
    max_l_dist = attrib(default=None)

    def __attrs_post_init__(self):
        self._check_params_valid()
        max_subs, max_ins, max_dels, max_l_dist = \
            self._normalize_params(*self.unpacked)
        object.__setattr__(self, 'max_substitutions', max_subs)
        object.__setattr__(self, 'max_insertions', max_ins)
        object.__setattr__(self, 'max_deletions', max_dels)
        object.__setattr__(self, 'max_l_dist', max_l_dist)

    @property
    def unpacked(self):
        return (
            self.max_substitutions,
            self.max_insertions,
            self.max_deletions,
            self.max_l_dist,
        )

    def _check_params_valid(self):
        if not all(x is None or (isinstance(x, int) and x >= 0)
                   for x in [
                       self.max_substitutions,
                       self.max_insertions,
                       self.max_deletions,
                       self.max_l_dist
                   ]):
            raise TypeError("All limits must be positive integers or None.")

        if self.max_l_dist is None:
            n_limits = (
                (1 if self.max_substitutions is not None else 0) +
                (1 if self.max_insertions is not None else 0) +
                (1 if self.max_deletions is not None else 0)
            )
            if n_limits < 3:
                if n_limits == 0:
                    raise ValueError('No limitations given!')
                elif self.max_substitutions is None:
                    raise ValueError('# substitutions must be limited!')
                elif self.max_insertions is None:
                    raise ValueError('# insertions must be limited!')
                elif self.max_deletions is None:
                    raise ValueError('# deletions must be limited!')

    @classmethod
    def _normalize_params(cls,
                          max_substitutions, max_insertions,
                          max_deletions, max_l_dist):
        maxes_sum = sum(
            x if x is not None else 1 << 29
            for x in [
                max_substitutions,
                max_insertions,
                max_deletions,
            ]
        )

        if max_l_dist is None:
            # replace max_l_dist with the sum of the other limits
            return (
                max_substitutions,
                max_insertions,
                max_deletions,
                maxes_sum,
            )
        else:
            def _normalize(param):
                return min(param, max_l_dist) if param is not None else max_l_dist
            return (
                _normalize(max_substitutions),
                _normalize(max_insertions),
                _normalize(max_deletions),
                min(max_l_dist, maxes_sum),
            )


def count_differences_with_maximum(sequence1, sequence2, max_differences):
    n_different = 0
    for item1, item2 in zip(sequence1, sequence2):
        if item1 != item2:
            n_different += 1
            if n_different == max_differences:
                return n_different
    return n_different

try:
    from fuzzysearch._common import count_differences_with_maximum_byteslike
except ImportError:
    pass
else:
    _count_differences_with_maximum = count_differences_with_maximum
    @wraps(_count_differences_with_maximum)
    def count_differences_with_maximum(sequence1, sequence2, max_differences):
        try:
            return count_differences_with_maximum_byteslike(sequence1,
                                                            sequence2,
                                                            max_differences)
        except TypeError:
            return _count_differences_with_maximum(sequence1, sequence2,
                                                   max_differences)


class GroupOfMatches(object):
    def __init__(self, match):
        assert match.start <= match.end
        self.start = match.start
        self.end = match.end
        self.matches = set([match])

    def is_match_in_group(self, match):
        return not (match.end <= self.start or match.start >= self.end)

    def add_match(self, match):
        self.matches.add(match)
        self.start = min(self.start, match.start)
        self.end = max(self.end, match.end)


def group_matches(matches):
    groups = []
    for match in matches:
        overlapping_groups = [g for g in groups if g.is_match_in_group(match)]
        if not overlapping_groups:
            groups.append(GroupOfMatches(match))
        elif len(overlapping_groups) == 1:
            overlapping_groups[0].add_match(match)
        else:
            new_group = GroupOfMatches(match)
            for group in overlapping_groups:
                for match in group.matches:
                    new_group.add_match(match)
            groups = [g for g in groups if g not in overlapping_groups]
            groups.append(new_group)

    return [group.matches for group in groups]


def get_best_match_in_group(group):
    """Get the longest match of those with the smallest distance."""
    return min(group, key=lambda match: (match.dist, -(match.end - match.start)))


def consolidate_overlapping_matches(matches):
    """Replace overlapping matches with a single, "best" match."""
    groups = group_matches(matches)
    best_matches = [get_best_match_in_group(group) for group in groups]
    return sorted(best_matches)


class FuzzySearchBase(object):
    """Abstract base class for fuzzy search classes"""
    @classmethod
    def search(cls, subsequence, sequence, search_params):
        raise NotImplementedError

    @classmethod
    def consolidate_matches(cls, matches):
        try:
            len(matches)
        except TypeError:
            return list(matches)
        else:
            return matches

    @classmethod
    def extra_items_for_chunked_search(cls, subsequence, search_params):
        raise NotImplementedError



"""A library for finding approximate subsequence matches.

Contains several implementations of fuzzy sub-sequence search functions. Such
functions find parts of a sequence which match a given sub-sequence up to a
given maximum Levenshtein distance.

The simplest use is via the find_near_matches utility function, which chooses
a suitable fuzzy search implementation based on the given parameters.

Example:
>>> find_near_matches('PATTERN', '---PATERN---', max_l_dist=1)
[Match(start=3, end=9, dist=1, matched='PATERN')]
"""
__author__ = 'Tal Einat'
__email__ = 'taleinat@gmail.com'
__version__ = '0.8.0'

__all__ = [
    'find_near_matches',
    'find_near_matches_in_file',
    'Match',
]

import io

from fuzzysearch.common import Match, LevenshteinSearchParams
from fuzzysearch.generic_search import GenericSearch
from fuzzysearch.levenshtein import LevenshteinSearch
from fuzzysearch.search_exact import ExactSearch
from fuzzysearch.substitutions_only import SubstitutionsOnlySearch

import attr


def find_near_matches(subsequence, sequence,
                      max_substitutions=None,
                      max_insertions=None,
                      max_deletions=None,
                      max_l_dist=None):
    """search for near-matches of subsequence in sequence

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the maximum allowed number of character substitutions
    * the maximum allowed number of new characters inserted
    * and the maximum allowed number of character deletions
    * the total number of substitutions, insertions and deletions
      (a.k.a. the Levenshtein distance)
    """
    search_params = LevenshteinSearchParams(max_substitutions,
                                            max_insertions,
                                            max_deletions,
                                            max_l_dist)
    search_class = choose_search_class(search_params)
    matches = search_class.search(subsequence, sequence, search_params)
    return search_class.consolidate_matches(matches)


def choose_search_class(search_params):
    max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked

    # if the limitations are so strict that only exact matches are allowed,
    # use search_exact()
    if max_l_dist == 0:
        return ExactSearch

    # if only substitutions are allowed, use find_near_matches_substitutions()
    elif max_insertions == 0 and max_deletions == 0:
        return SubstitutionsOnlySearch

    # if it is enough to just take into account the maximum Levenshtein
    # distance, use find_near_matches_levenshtein()
    elif max_l_dist <= min(
        (max_substitutions if max_substitutions is not None else (1 << 29)),
        (max_insertions if max_insertions is not None else (1 << 29)),
        (max_deletions if max_deletions is not None else (1 << 29)),
    ):
        return LevenshteinSearch

    # if none of the special cases above are met, use the most generic version
    else:
        return GenericSearch


def find_near_matches_in_file(subsequence, sequence_file,
                              max_substitutions=None,
                              max_insertions=None,
                              max_deletions=None,
                              max_l_dist=None,
                              _chunk_size=2**20):
    """search for near-matches of subsequence in a file

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the maximum allowed number of character substitutions
    * the maximum allowed number of new characters inserted
    * and the maximum allowed number of character deletions
    * the total number of substitutions, insertions and deletions
      (a.k.a. the Levenshtein distance)
    """
    search_params = LevenshteinSearchParams(max_substitutions,
                                            max_insertions,
                                            max_deletions,
                                            max_l_dist)
    search_class = choose_search_class(search_params)

    if (
            'b' in getattr(sequence_file, 'mode', '')
            or
            isinstance(sequence_file, io.RawIOBase)
    ):
        matches = _search_binary_file(subsequence,
                                      sequence_file,
                                      search_params,
                                      search_class,
                                      _chunk_size=_chunk_size)
    else:
        matches = _search_unicode_file(subsequence,
                                       sequence_file,
                                       search_params,
                                       search_class,
                                       _chunk_size=_chunk_size)

    return search_class.consolidate_matches(matches)


def _search_binary_file(subsequence, sequence_file, search_params, search_class,
                        _chunk_size):
    if not subsequence:
        raise ValueError('subsequence must not be empty')

    CHUNK_SIZE = _chunk_size
    keep_bytes = (
        len(subsequence) - 1 +
        search_class.extra_items_for_chunked_search(subsequence, search_params)
    )

    # To allocate memory only once, we'll use a pre-allocated bytearray and
    # file.readinto().  Furthermore, since we'll need to keep part of each
    # chunk along with the next chunk, we'll use a memoryview of the bytearray
    # to move data around within a single block of memory and thus avoid
    # allocations.
    chunk_bytes = bytearray(CHUNK_SIZE)
    chunk_memview = memoryview(chunk_bytes)

    # The search will be done with bytearray objects.  Note that in Python 2,
    # getting an item from a bytes object returns a string (rather than an
    # int as in Python 3), so we explicitly convert the sub-sequence to a
    # bytearray in case it is a bytes/str object.
    subseq_bytearray = bytearray(subsequence)

    n_read = sequence_file.readinto(chunk_memview)
    offset = 0
    chunk_len = n_read
    while n_read:
        search_bytes = chunk_bytes if chunk_len == CHUNK_SIZE else chunk_bytes[:chunk_len]
        for match in search_class.search(subseq_bytearray, search_bytes, search_params):
            yield attr.evolve(match,
                              start=match.start + offset,
                              end=match.end + offset)

        if keep_bytes > 0:
            n_to_keep = min(keep_bytes, chunk_len)
            chunk_memview[:n_to_keep] = chunk_memview[chunk_len - n_to_keep:chunk_len]
        else:
            n_to_keep = 0
        offset += chunk_len - n_to_keep
        n_read = sequence_file.readinto(chunk_memview[n_to_keep:])
        chunk_len = n_to_keep + n_read


def _search_unicode_file(subsequence, sequence_file, search_params, search_class,
                         _chunk_size):
    if not subsequence:
        raise ValueError('subsequence must not be empty')

    CHUNK_SIZE = _chunk_size
    keep_chars = (
        len(subsequence) - 1 +
        search_class.extra_items_for_chunked_search(subsequence, search_params)
    )

    chunk = sequence_file.read(CHUNK_SIZE)
    offset = 0
    while chunk:
        for match in search_class.search(subsequence, chunk, search_params):
            yield attr.evolve(match,
                              start=match.start + offset,
                              end=match.end + offset)

        n_to_keep = min(keep_chars, len(chunk))
        offset += len(chunk) - n_to_keep
        if n_to_keep:
            chunk = chunk[-n_to_keep:] + sequence_file.read(CHUNK_SIZE)
            if len(chunk) == n_to_keep:
                break
        else:
            chunk = sequence_file.read(CHUNK_SIZE)



import unittest

from fuzzysearch.common import Match, LevenshteinSearchParams
from fuzzysearch.no_deletions import _expand, \
    find_near_matches_no_deletions_ngrams
from tests.test_substitutions_only import TestFindNearMatchesSubstitionsNgrams


def fnm_nodels_ngrams(sequence, subsequence, max_substitutions, max_insertions, max_l_dist=None):
    return find_near_matches_no_deletions_ngrams(
        sequence, subsequence, LevenshteinSearchParams(
            max_substitutions, max_insertions, 0, max_l_dist,
        )
    )


class TestExpand(unittest.TestCase):
    def test_identical(self):
        self.assertEqual(_expand('abc', 'abc', 0, 0, 0), [(0, 0)])

    def test_startswith(self):
        self.assertEqual(_expand('abc', 'abcdef', 0, 0, 0), [(0, 0)])
        self.assertEqual(_expand('abc', 'abcdef', 1, 0, 1), [(0, 0)])
        self.assertEqual(_expand('abc', 'abcdef', 2, 0, 2), [(0, 0)])
        self.assertEqual(_expand('abc', 'abcdef', 0, 1, 1), [(0, 0)])
        self.assertEqual(_expand('abc', 'abcdef', 0, 2, 2), [(0, 0)])
        self.assertEqual(_expand('abc', 'abcdef', 1, 1, 1), [(0, 0)])
        self.assertEqual(_expand('abc', 'abcdef', 1, 1, 2), [(0, 0)])
        self.assertEqual(_expand('abc', 'abcdef', 2, 2, 2), [(0, 0)])

    def test_one_missing(self):
        # first item missing
        self.assertEqual(_expand('abcd', 'bcd---', 0, 1, 1), [])
        self.assertEqual(_expand('abcd', 'bcd---', 1, 0, 1), [])
        self.assertEqual(_expand('abcd', 'bcd---', 1, 1, 2), [])

        # second item missing
        self.assertEqual(_expand('abcd', 'acd---', 0, 1, 1), [])
        self.assertEqual(_expand('abcd', 'acd---', 1, 0, 1), [])
        self.assertEqual(_expand('abcd', 'acd---', 1, 1, 2), [])

        # last item missing
        self.assertEqual(_expand('abcd', 'abc---', 0, 1, 1), [])
        self.assertEqual(_expand('abcd', 'abc---', 1, 0, 1), [(1, 0)])
        self.assertEqual(_expand('abcd', 'abc---', 1, 1, 2), [(1, 0)])

    def test_no_result(self):
        self.assertEqual(_expand('abc', 'def', 0, 0, 0), [])
        self.assertEqual(_expand('abc', 'defg', 1, 1, 1), [])
        self.assertEqual(_expand('abc', 'defg', 1, 1, 2), [])

    def test_one_extra(self):
        # extra first item
        self.assertEqual(_expand('bcd', 'abcd', 0, 0, 0), [])
        self.assertEqual(_expand('bcd', 'abcd', 0, 1, 1), [(0, 1)])

        # extra third item
        self.assertEqual(_expand('abd', 'abcd', 0, 0, 0), [])
        self.assertEqual(_expand('abd', 'abcd', 0, 1, 1), [(0, 1)])

        # extra last item
        self.assertEqual(_expand('abc', 'abcd', 0, 0, 0), [(0, 0)])
        self.assertEqual(_expand('abc', 'abcd', 0, 1, 1), [(0, 0)])

    def test_insert_and_substitute(self):
        self.assertEqual(_expand('abcdefg', 'abc-def----', 1, 1, 2), [(1, 1)])
        self.assertEqual(_expand('abcdefg', 'abc-def----', 1, 1, 1), [])
        self.assertEqual(_expand('abcdefg', 'abc-def----', 1, 0, 1), [])
        self.assertEqual(_expand('abcdefg', 'abc-def----', 0, 1, 1), [])

    def test_double_first_item(self):
        self.assertEqual(_expand('abc', 'aabc', 1, 1, 1), [(0, 1)])

    def test_two_insertions(self):
        self.assertEqual(_expand('abc', 'a--bc', 0, 2, 2), [(0, 2)])
        self.assertEqual(_expand('abc', 'a--bc', 2, 0, 2), [(2, 0)])
        self.assertEqual(_expand('abc', 'a--bc', 2, 2, 2), [(2, 0), (0, 2)])
        self.assertEqual(_expand('abc', 'a--bc', 1, 1, 2), [])


class TestFindNearMatchesNoDeletionsNgramsAsNoSubstituions(
    TestFindNearMatchesSubstitionsNgrams, unittest.TestCase):
    def search(self, subsequence, sequence, max_subs):
        if max_subs >= len(subsequence):
            self.skipTest("avoiding calling fnm_no_deletions_ngrams() " +
                          "with max_subs >= len(subsequence)")
        return fnm_nodels_ngrams(subsequence, sequence, max_subs, 0)


class TestFindNearMatchesNoDeletionsNgrams(unittest.TestCase):
    def test_one_sub_one_ins(self):
        sequence = 'abcdefghij'
        pattern = 'bceXghi'
        expected_match = Match(start=1, end=9, dist=2, matched=sequence[1:9])
        self.assertEqual(fnm_nodels_ngrams(pattern, sequence, 0, 0, 0), [])
        self.assertEqual(fnm_nodels_ngrams(pattern, sequence, 0, 1, 2), [])
        self.assertEqual(fnm_nodels_ngrams(pattern, sequence, 1, 0, 2), [])
        self.assertEqual(fnm_nodels_ngrams(pattern, sequence, 1, 1, 1), [])
        self.assertEqual(
            fnm_nodels_ngrams(pattern, sequence, 1, 1, 2),
            [expected_match],
        )

    def test_two_extra(self):
        sequence = '--abc--de--'
        pattern = 'abcde'

        self.assertEqual(
            fnm_nodels_ngrams(pattern, sequence, 0, 2, 2),
            [Match(start=2, end=9, dist=2, matched=sequence[2:9])],
        )

        self.assertEqual(
            fnm_nodels_ngrams(pattern, sequence, 2, 0, 2),
            [Match(start=2, end=7, dist=2, matched=sequence[2:7])],
        )

        self.assertEqual(
            fnm_nodels_ngrams(pattern, sequence, 2, 2, 2),
            [Match(start=2, end=7, dist=2, matched=sequence[2:7]),
             Match(start=2, end=9, dist=2, matched=sequence[2:9])],
        )



from fuzzysearch.common import Match
from fuzzysearch.search_exact import search_exact


__all__ = ['find_near_matches_levenshtein_ngrams']


def _expand(subsequence, sequence, max_l_dist):
    """Expand a partial match of a Levenstein search.

    An expansion must begin at the beginning of the sequence, which makes
    this much simpler than a full search, and allows for greater optimization.
    """
    # If given a long sub-sequence and relatively small max distance,
    # use a more complex algorithm better optimized for such cases.
    if len(subsequence) > max(max_l_dist * 2, 10):
        return _expand_long(subsequence, sequence, max_l_dist)
    else:
        return _expand_short(subsequence, sequence, max_l_dist)


def _py_expand_short(subsequence, sequence, max_l_dist):
    """Straightforward implementation of partial match expansion."""
    # The following diagram shows the score calculation step.
    #
    # Each new score is the minimum of:
    #  * a OR a + 1 (substitution, if needed)
    #  * b + 1 (deletion, i.e. skipping a sequence character)
    #  * c + 1 (insertion, i.e. skipping a sub-sequence character)
    #
    # a -- +1 -> c
    #
    # |  \       |
    # |   \      |
    # +1  +1?    +1
    # |      \   |
    # v       \  v
    #
    # b -- +1 -> scores[subseq_index]

    subseq_len = len(subsequence)
    if subseq_len == 0:
        return (0, 0)

    # Initialize the scores array with values for just skipping sub-sequence
    # chars.
    scores = list(range(1, subseq_len + 1))

    min_score = subseq_len
    min_score_idx = -1

    for seq_index, char in enumerate(sequence):
        # calculate scores, one for each character in the sub-sequence
        a = seq_index
        c = a + 1
        for subseq_index in range(subseq_len):
            b = scores[subseq_index]
            c = scores[subseq_index] = min(
                a + (char != subsequence[subseq_index]),
                b + 1,
                c + 1,
            )
            a = b

        # keep the minimum score found for matches of the entire sub-sequence
        if c <= min_score:
            min_score = c
            min_score_idx = seq_index

        # bail early when it is impossible to find a better expansion
        elif min(scores) >= min_score:
            break

    return (min_score, min_score_idx + 1) if min_score <= max_l_dist else (None, None)


def _py_expand_long(subsequence, sequence, max_l_dist):
    """Partial match expansion, optimized for long sub-sequences."""
    # The additional optimization in this version is to limit the part of
    # the sub-sequence inspected for each sequence character.  The start and
    # end of the iteration are limited to the range where the scores are
    # smaller than the maximum allowed distance.  Additionally, once a good
    # expansion has been found, the range is further reduced to where the
    # scores are smaller than the score of the best expansion found so far.

    subseq_len = len(subsequence)
    if subseq_len == 0:
        return (0, 0)

    # Initialize the scores array with values for just skipping sub-sequence
    # chars.
    scores = list(range(1, subseq_len + 1))

    min_score = subseq_len
    min_score_idx = -1
    max_good_score = max_l_dist
    new_needle_idx_range_start = 0
    new_needle_idx_range_end = subseq_len - 1

    for seq_index, char in enumerate(sequence):
        # calculate scores, one for each character in the sub-sequence
        needle_idx_range_start = new_needle_idx_range_start
        needle_idx_range_end = min(subseq_len, new_needle_idx_range_end + 1)

        a = seq_index
        c = a + 1

        if c <= max_good_score:
            new_needle_idx_range_start = 0
            new_needle_idx_range_end = 0
        else:
            new_needle_idx_range_start = None
            new_needle_idx_range_end = -1

        for subseq_index in range(needle_idx_range_start, needle_idx_range_end):
            b = scores[subseq_index]
            c = scores[subseq_index] = min(
                a + (char != subsequence[subseq_index]),
                b + 1,
                c + 1,
            )
            a = b

            if c <= max_good_score:
                if new_needle_idx_range_start is None:
                    new_needle_idx_range_start = subseq_index
                new_needle_idx_range_end = max(
                    new_needle_idx_range_end,
                    subseq_index + 1 + (max_good_score - c),
                )

        # bail early when it is impossible to find a better expansion
        if new_needle_idx_range_start is None:
            break

        # keep the minimum score found for matches of the entire sub-sequence
        if needle_idx_range_end == subseq_len and c <= min_score:
            min_score = c
            min_score_idx = seq_index
            if min_score < max_good_score:
                max_good_score = min_score

    return (min_score, min_score_idx + 1) if min_score <= max_l_dist else (None, None)


try:
    from fuzzysearch._levenshtein_ngrams import (
        c_expand_short as _c_expand_short,
        c_expand_long as _c_expand_long,
    )
except ImportError:
    _expand_short = _py_expand_short
    _expand_long = _py_expand_long
else:
    _expand_short = _c_expand_short
    _expand_long = _c_expand_long


def find_near_matches_levenshtein_ngrams(subsequence, sequence, max_l_dist):
    subseq_len = len(subsequence)
    seq_len = len(sequence)

    ngram_len = subseq_len // (max_l_dist + 1)
    if ngram_len == 0:
        raise ValueError('the subsequence length must be greater than max_l_dist')

    def make_match(start, end, dist):
        return Match(start, end, dist, matched=sequence[start:end])

    for ngram_start in range(0, subseq_len - ngram_len + 1, ngram_len):
        ngram_end = ngram_start + ngram_len
        subseq_before_reversed = subsequence[:ngram_start][::-1]
        subseq_after = subsequence[ngram_end:]
        start_index = max(0, ngram_start - max_l_dist)
        end_index = min(seq_len, seq_len - subseq_len + ngram_end + max_l_dist)
        for index in search_exact(subsequence[ngram_start:ngram_end], sequence, start_index, end_index):
            # try to expand left and/or right according to n_ngram
            dist_right, right_expand_size = _expand(
                subseq_after,
                sequence[index + ngram_len:index - ngram_start + subseq_len + max_l_dist],
                max_l_dist,
            )
            if dist_right is None:
                continue
            dist_left, left_expand_size = _expand(
                subseq_before_reversed,
                sequence[max(0, index - ngram_start - (max_l_dist - dist_right)):index][::-1],
                max_l_dist - dist_right,
            )
            if dist_left is None:
                continue
            assert dist_left + dist_right <= max_l_dist

            yield make_match(
                start=index - left_expand_size,
                end=index + ngram_len + right_expand_size,
                dist=dist_left + dist_right,
            )



from collections import deque, defaultdict
from itertools import islice
from functools import wraps

from fuzzysearch.common import FuzzySearchBase, Match, \
    count_differences_with_maximum, get_best_match_in_group, group_matches
from fuzzysearch.search_exact import search_exact


def _check_arguments(subsequence, sequence, max_substitutions):
    if not subsequence:
        raise ValueError('Given subsequence is empty!')

    if max_substitutions is None or max_substitutions < 0:
        raise ValueError('Maximum number of substitutions must be >= 0!')


def has_near_match_substitutions(subsequence, sequence, max_substitutions):
    _check_arguments(subsequence, sequence, max_substitutions)

    if max_substitutions == 0:
        for start_index in search_exact(subsequence, sequence):
            return True
        return False

    elif len(subsequence) // (max_substitutions + 1) >= 3:
        return has_near_match_substitutions_ngrams(
            subsequence, sequence, max_substitutions,
        )

    else:
        return has_near_match_substitutions_lp(
            subsequence, sequence, max_substitutions,
        )


def find_near_matches_substitutions(subsequence, sequence, max_substitutions):
    """Find near-matches of the subsequence in the sequence.

    This chooses a suitable fuzzy search implementation according to the given
    parameters.

    Returns a list of fuzzysearch.Match objects describing the matching parts
    of the sequence.
    """
    _check_arguments(subsequence, sequence, max_substitutions)

    if max_substitutions == 0:
        return [
            Match(start_index, start_index + len(subsequence), 0,
                  sequence[start_index:start_index + len(subsequence)])
            for start_index in search_exact(subsequence, sequence)
        ]

    elif len(subsequence) // (max_substitutions + 1) >= 3:
        return find_near_matches_substitutions_ngrams(
            subsequence, sequence, max_substitutions,
        )

    else:
        return find_near_matches_substitutions_lp(
            subsequence, sequence, max_substitutions,
        )


def find_near_matches_substitutions_lp(subsequence, sequence,
                                       max_substitutions):
    """search for near-matches of subsequence in sequence

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the number of character substitutions must be less than max_substitutions
    * no deletions or insertions are allowed
    """
    _check_arguments(subsequence, sequence, max_substitutions)

    return list(_find_near_matches_substitutions_lp(subsequence, sequence,
                                                    max_substitutions))


def _find_near_matches_substitutions_lp(subsequence, sequence,
                                        max_substitutions):
    # simple optimization: prepare some often used things in advance
    _SUBSEQ_LEN = len(subsequence)
    _SUBSEQ_LEN_MINUS_ONE = _SUBSEQ_LEN - 1

    def make_match(start, end, dist):
        return Match(start, end, dist, matched=sequence[start:end])

    # prepare quick lookup of where a character appears in the subsequence
    char_indexes_in_subsequence = defaultdict(list)
    for (index, char) in enumerate(subsequence):
        char_indexes_in_subsequence[char].append(index)

    # we'll iterate over the sequence once, but the iteration is split into two
    # for loops; therefore we prepare an iterator in advance which will be used
    # in both of the loops
    sequence_enum_iter = enumerate(sequence)

    # We'll count the number of matching characters assuming various attempted
    # alignments of the subsequence to the sequence. At any point in the
    # sequence there will be N such alignments to update. We'll keep
    # these in a "circular array" (a.k.a. a ring) which we'll rotate after each
    # iteration to re-align the indexing.

    # Initialize the candidate counts by iterating over the first N-1 items in
    # the sequence. No possible matches in this step!
    candidates = deque([0], maxlen=_SUBSEQ_LEN)
    for (index, char) in islice(sequence_enum_iter, _SUBSEQ_LEN_MINUS_ONE):
        for subseq_index in [idx for idx in char_indexes_in_subsequence[char] if idx <= index]:
            candidates[subseq_index] += 1
        candidates.appendleft(0)

    # From the N-th item onwards, we'll update the candidate counts exactly as
    # above, and additionally check if the part of the sequence whic began N-1
    # items before the current index was a near enough match to the given
    # sub-sequence.
    for (index, char) in sequence_enum_iter:
        for subseq_index in char_indexes_in_subsequence[char]:
            candidates[subseq_index] += 1

        # rotate the ring of candidate counts
        candidates.rotate(1)
        # fetch the count for the candidate which started N-1 items ago
        n_substitutions = _SUBSEQ_LEN - candidates[0]
        # set the count for the next index to zero
        candidates[0] = 0

        # if the candidate had few enough mismatches, yield a match
        if n_substitutions <= max_substitutions:
            yield make_match(
                start=index - _SUBSEQ_LEN_MINUS_ONE,
                end=index + 1,
                dist=n_substitutions,
            )


def has_near_match_substitutions_lp(subsequence, sequence, max_substitutions):
    _check_arguments(subsequence, sequence, max_substitutions)

    for match in _find_near_matches_substitutions_lp(subsequence, sequence,
                                                     max_substitutions):
        return True
    return False


def find_near_matches_substitutions_ngrams(subsequence, sequence,
                                           max_substitutions):
    """search for near-matches of subsequence in sequence

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the number of character substitutions must be less than max_substitutions
    * no deletions or insertions are allowed
    """
    _check_arguments(subsequence, sequence, max_substitutions)

    match_starts = set()
    matches = []
    for match in _find_near_matches_substitutions_ngrams(subsequence, sequence,
                                                         max_substitutions):
        if match.start not in match_starts:
            match_starts.add(match.start)
            matches.append(match)
    return sorted(matches, key=lambda match: match.start)


def _find_near_matches_substitutions_ngrams(subsequence, sequence,
                                            max_substitutions):
    subseq_len = len(subsequence)
    seq_len = len(sequence)

    def make_match(start, end, dist):
        return Match(start, end, dist, matched=sequence[start:end])

    ngram_len = subseq_len // (max_substitutions + 1)
    if ngram_len == 0:
        raise ValueError(
            "The subsequence's length must be greater than max_substitutions!"
        )

    for ngram_start in range(0, len(subsequence) - ngram_len + 1, ngram_len):
        ngram_end = ngram_start + ngram_len
        subseq_before = subsequence[:ngram_start]
        subseq_after = subsequence[ngram_end:]
        for index in search_exact(
                subsequence[ngram_start:ngram_end], sequence,
                ngram_start, seq_len - (subseq_len - ngram_end),
        ):
            n_substitutions = 0
            seq_before = sequence[index - ngram_start:index]
            if subseq_before != seq_before:
                n_substitutions += count_differences_with_maximum(
                    seq_before, subseq_before,
                    max_substitutions - n_substitutions + 1)
                if n_substitutions > max_substitutions:
                    continue

            seq_after = sequence[index + ngram_len:index - ngram_start + subseq_len]
            if subseq_after != seq_after:
                if n_substitutions == max_substitutions:
                    continue
                n_substitutions += count_differences_with_maximum(
                    seq_after, subseq_after,
                    max_substitutions - n_substitutions + 1)
                if n_substitutions > max_substitutions:
                    continue

            yield make_match(
                start=index - ngram_start,
                end=index - ngram_start + subseq_len,
                dist=n_substitutions,
            )


def has_near_match_substitutions_ngrams(subsequence, sequence,
                                        max_substitutions):
    """search for near-matches of subsequence in sequence

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the number of character substitutions must be less than max_substitutions
    * no deletions or insertions are allowed
    """
    _check_arguments(subsequence, sequence, max_substitutions)

    for match in _find_near_matches_substitutions_ngrams(subsequence, sequence,
                                                         max_substitutions):
        return True
    return False


try:
    from fuzzysearch._substitutions_only import \
        substitutions_only_has_near_matches_ngrams_byteslike, \
        substitutions_only_find_near_matches_ngrams_byteslike as \
            _subs_only_fnm_ngram_byteslike
except ImportError:
    pass
else:
    py_has_near_match_substitutions_ngrams = has_near_match_substitutions_ngrams
    @wraps(py_has_near_match_substitutions_ngrams)
    def has_near_match_substitutions_ngrams(subsequence, sequence,
                                            max_substitutions):
        try:
            return substitutions_only_has_near_matches_ngrams_byteslike(
                subsequence, sequence, max_substitutions)
        except (TypeError, UnicodeEncodeError):
            return py_has_near_match_substitutions_ngrams(
                subsequence, sequence, max_substitutions)

    py_find_near_matches_substitutions_ngrams = \
        find_near_matches_substitutions_ngrams
    @wraps(py_find_near_matches_substitutions_ngrams)
    def find_near_matches_substitutions_ngrams(subsequence, sequence,
                                               max_substitutions):
        try:
            results = _subs_only_fnm_ngram_byteslike(
                subsequence, sequence, max_substitutions)
        except (TypeError, UnicodeEncodeError):
            pass
        else:
            matches = [
                Match(
                    index,
                    index + len(subsequence),
                    count_differences_with_maximum(
                        sequence[index:index+len(subsequence)],
                        subsequence,
                        max_substitutions + 1,
                    ),
                    matched=sequence[index:index + len(subsequence)],
                )
                for index in results
            ]
            return [
                get_best_match_in_group(group)
                for group in group_matches(matches)
            ]

        return py_find_near_matches_substitutions_ngrams(
            subsequence, sequence, max_substitutions)


class SubstitutionsOnlySearch(FuzzySearchBase):
    @classmethod
    def search(cls, subsequence, sequence, search_params):
        actual_max_subs = min(
            x for x in [search_params.max_l_dist,
                        search_params.max_substitutions]
            if x is not None
        )
        return find_near_matches_substitutions(subsequence, sequence,
                                               actual_max_subs)

    @classmethod
    def extra_items_for_chunked_search(cls, subsequence, search_params):
        return 0



import unittest

from tests.compat import b
from tests.utils import skip_if_arguments_arent_byteslike
from tests.test_levenshtein import TestFindNearMatchesLevenshteinBase
from fuzzysearch.common import Match, get_best_match_in_group, group_matches, LevenshteinSearchParams
from tests.test_substitutions_only import TestSubstitionsOnlyBase
from tests.test_generic_search import TestGenericSearchBase


try:
    from fuzzysearch._generic_search import \
        c_find_near_matches_generic_linear_programming as c_fnm_generic_lp, \
        c_find_near_matches_generic_ngrams as c_fnm_generic_ngrams
except ImportError:
    pass
else:
    class TestGenericSearchLpAsLevenshtein(TestFindNearMatchesLevenshteinBase,
                                           unittest.TestCase):
        @skip_if_arguments_arent_byteslike
        def search(self, subsequence, sequence, max_l_dist):
            return [
                get_best_match_in_group(group)
                for group in group_matches(
                    c_fnm_generic_lp(subsequence,
                                     sequence,
                                     LevenshteinSearchParams(max_l_dist, max_l_dist, max_l_dist, max_l_dist))
                )
            ]

    class TestGenericSearchNgramsAsLevenshtein(
        TestFindNearMatchesLevenshteinBase, unittest.TestCase):
        @skip_if_arguments_arent_byteslike
        def search(self, subsequence, sequence, max_l_dist):
            if max_l_dist >= len(subsequence):
                self.skipTest("avoiding calling c_fnm_generic_ngrams() " +
                              "with max_l_dist >= len(subsequence)")
            return [
                get_best_match_in_group(group)
                for group in group_matches(
                    c_fnm_generic_ngrams(subsequence,
                                         sequence,
                                         LevenshteinSearchParams(max_l_dist, max_l_dist, max_l_dist, max_l_dist))
                )
            ]


    class TestGenericSearchLpAsSubstitutionsOnly(TestSubstitionsOnlyBase,
                                                 unittest.TestCase):
        @skip_if_arguments_arent_byteslike
        def search(self, subsequence, sequence, max_subs):
            return list(
                c_fnm_generic_lp(subsequence,
                                 sequence,
                                 LevenshteinSearchParams(max_subs, 0, 0, max_subs))
            )

        def expectedOutcomes(self, search_results, expected_outcomes,
                             *args, **kwargs):
            return self.assertEqual(search_results, expected_outcomes,
                                    *args, **kwargs)


    class TestGenericSearchNgramsAsSubstitutionsOnly(TestSubstitionsOnlyBase,
                                                     unittest.TestCase):
        @skip_if_arguments_arent_byteslike
        def search(self, subsequence, sequence, max_subs):
            if max_subs >= len(subsequence):
                self.skipTest("avoiding calling c_fnm_generic_ngrams() " +
                              "with max_subs >= len(subsequence)")
            return [
                get_best_match_in_group(group)
                for group in group_matches(
                    c_fnm_generic_ngrams(subsequence,
                                         sequence,
                                         LevenshteinSearchParams(max_subs, 0, 0, max_subs))
                )
        ]

        def expectedOutcomes(self, search_results, expected_outcomes,
                             *args, **kwargs):
            best_from_grouped_results = [
                get_best_match_in_group(group)
                for group in group_matches(search_results)
            ]
            best_from_grouped_exepected_outcomes = [
                get_best_match_in_group(group)
                for group in group_matches(expected_outcomes)
            ]
            return self.assertEqual(best_from_grouped_results,
                                    best_from_grouped_exepected_outcomes,
                                    *args, **kwargs)


    class TestGenericSearchLp(TestGenericSearchBase, unittest.TestCase):
        @skip_if_arguments_arent_byteslike
        def search(self, pattern, sequence, max_subs, max_ins, max_dels,
                   max_l_dist=None):
            return list(c_fnm_generic_lp(pattern,
                                         sequence,
                                         LevenshteinSearchParams(
                                             max_subs, max_ins,
                                             max_dels, max_l_dist,
                                         )))

        def expectedOutcomes(self, search_result, expected_outcomes,
                             *args, **kwargs):
            self.assertEqual(search_result, expected_outcomes, *args, **kwargs)

        def test_double_first_item_two_results(self):
            self.assertEqual(
                self.search(b('def'), b('abcddefg'), 0, 1, 0),
                [Match(start=3, end=7, dist=1, matched=b('ddef')),
                 Match(start=4, end=7, dist=0, matched=b('def'))],
            )

        def test_missing_second_item_complex(self):
            self.assertEqual(
                self.search(b('bde'), b('abcdefg'), 1, 1, 1, 1),
                [Match(start=1, end=5, dist=1, matched=b('bcde')),
                 Match(start=2, end=5, dist=1, matched=b('cde')),
                 Match(start=3, end=5, dist=1, matched=b('de'))],
            )

            self.assertTrue(
                {
                    Match(start=1, end=5, dist=1, matched=b('bcde')),
                    Match(start=2, end=5, dist=1, matched=b('cde')),
                    Match(start=3, end=5, dist=1, matched=b('de')),
                    Match(start=2, end=5, dist=2, matched=b('bcd')),
                }.issubset(set(
                    self.search(b('bde'), b('abcdefg'), 1, 1, 1, 3),
                ))
            )

    class TestGenericSearchNgrams(TestGenericSearchBase, unittest.TestCase):
        @skip_if_arguments_arent_byteslike
        def search(self, pattern, sequence, max_subs, max_ins, max_dels,
                   max_l_dist=None):
            return [
                get_best_match_in_group(group)
                for group in group_matches(
                    c_fnm_generic_ngrams(pattern,
                                         sequence,
                                         LevenshteinSearchParams(
                                             max_subs, max_ins,
                                             max_dels, max_l_dist,
                                         ))
                )
            ]

        def expectedOutcomes(self, search_result, expected_outcomes,
                             *args, **kwargs):
            best_from_groups = [
                get_best_match_in_group(group)
                for group in group_matches(search_result)
            ]
            self.assertEqual(search_result, best_from_groups, *args, **kwargs)

        def test_missing_second_item_complex(self):
            self.assertTrue(
                set(self.search(b('bde'), b('abcdefg'), 1, 1, 1, 1)).issubset([
                    Match(start=1, end=5, dist=1, matched=b('bcde')),
                    Match(start=2, end=5, dist=1, matched=b('cde')),
                    Match(start=3, end=5, dist=1, matched=b('de')),
                ])
            )



from collections import namedtuple
from functools import wraps

import attr

from fuzzysearch.common import FuzzySearchBase, Match, \
    consolidate_overlapping_matches
from fuzzysearch.search_exact import search_exact


__all__ = [
    'find_near_matches_generic',
    'find_near_matches_generic_linear_programming',
    'find_near_matches_generic_ngrams',
    'has_near_match_generic_ngrams',
]


GenericSearchCandidate = namedtuple(
    'GenericSearchCandidate',
    ['start', 'subseq_index', 'l_dist', 'n_subs', 'n_ins', 'n_dels'],
)


def find_near_matches_generic(subsequence, sequence, search_params):
    """search for near-matches of subsequence in sequence

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the maximum allowed number of character substitutions
    * the maximum allowed number of new characters inserted
    * and the maximum allowed number of character deletions
    * the total number of substitutions, insertions and deletions
    """
    if not subsequence:
        raise ValueError('Given subsequence is empty!')

    # if the limitations are so strict that only exact matches are allowed,
    # use search_exact()
    if search_params.max_l_dist == 0:
        return [
            Match(start_index, start_index + len(subsequence), 0,
                  matched=sequence[start_index:start_index + len(subsequence)])
            for start_index in search_exact(subsequence, sequence)
        ]

    # if the n-gram length would be at least 3, use the n-gram search method
    elif len(subsequence) // (search_params.max_l_dist + 1) >= 3:
        return find_near_matches_generic_ngrams(subsequence, sequence, search_params)

    # use the linear programming search method
    else:
        return find_near_matches_generic_linear_programming(subsequence, sequence, search_params)


def _find_near_matches_generic_linear_programming(subsequence, sequence, search_params):
    """search for near-matches of subsequence in sequence

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the maximum allowed number of character substitutions
    * the maximum allowed number of new characters inserted
    * and the maximum allowed number of character deletions
    * the total number of substitutions, insertions and deletions
    """
    if not subsequence:
        raise ValueError('Given subsequence is empty!')

    max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked

    # optimization: prepare some often used things in advance
    subseq_len = len(subsequence)

    def make_match(start, end, dist):
        return Match(start, end, dist, matched=sequence[start:end])

    candidates = []
    for index, char in enumerate(sequence):
        candidates.append(GenericSearchCandidate(index, 0, 0, 0, 0, 0))
        new_candidates = []

        for cand in candidates:
            # if this sequence char is the candidate's next expected char
            if char == subsequence[cand.subseq_index]:
                # if reached the end of the subsequence, return a match
                if cand.subseq_index + 1 == subseq_len:
                    yield make_match(cand.start, index + 1, cand.l_dist)
                # otherwise, update the candidate's subseq_index and keep it
                else:
                    new_candidates.append(cand._replace(
                        subseq_index=cand.subseq_index + 1,
                    ))

            # if this sequence char is *not* the candidate's next expected char
            else:
                # we can try skipping a sequence or sub-sequence char (or both),
                # unless this candidate has already skipped the maximum allowed
                # number of characters
                if cand.l_dist == max_l_dist:
                    continue

                if cand.n_ins < max_insertions:
                    # add a candidate skipping a sequence char
                    new_candidates.append(cand._replace(
                        n_ins=cand.n_ins + 1,
                        l_dist=cand.l_dist + 1,
                    ))

                if cand.subseq_index + 1 < subseq_len:
                    if cand.n_subs < max_substitutions:
                        # add a candidate skipping both a sequence char and a
                        # subsequence char
                        new_candidates.append(cand._replace(
                            n_subs=cand.n_subs + 1,
                            subseq_index=cand.subseq_index + 1,
                            l_dist=cand.l_dist + 1,
                        ))
                    elif cand.n_dels < max_deletions and cand.n_ins < max_insertions:
                        # add a candidate skipping both a sequence char and a
                        # subsequence char
                        new_candidates.append(cand._replace(
                            n_ins=cand.n_ins + 1,
                            n_dels=cand.n_dels + 1,
                            subseq_index=cand.subseq_index + 1,
                            l_dist=cand.l_dist + 1,
                        ))
                else:
                    # cand.subseq_index == _subseq_len - 1
                    if (
                            cand.n_subs < max_substitutions or
                            (
                                cand.n_dels < max_deletions and
                                cand.n_ins < max_insertions
                            )
                    ):
                        yield make_match(cand.start, index + 1, cand.l_dist + 1)

                # try skipping subsequence chars
                for n_skipped in range(1, min(max_deletions - cand.n_dels, max_l_dist - cand.l_dist) + 1):
                    # if skipping n_dels sub-sequence chars reaches the end
                    # of the sub-sequence, yield a match
                    if cand.subseq_index + n_skipped == subseq_len:
                        yield make_match(cand.start, index,
                                         cand.l_dist + n_skipped)
                        break
                    # otherwise, if skipping n_skipped sub-sequence chars
                    # reaches a sub-sequence char identical to this sequence
                    # char ...
                    elif subsequence[cand.subseq_index + n_skipped] == char:
                        # if this is the last char of the sub-sequence, yield
                        # a match
                        if cand.subseq_index + n_skipped + 1 == subseq_len:
                            yield make_match(cand.start, index,
                                             cand.l_dist + n_skipped)
                        # otherwise add a candidate skipping n_skipped
                        # subsequence chars
                        else:
                            new_candidates.append(cand._replace(
                                n_dels=cand.n_dels + n_skipped,
                                subseq_index=cand.subseq_index + 1 + n_skipped,
                                l_dist=cand.l_dist + n_skipped,
                            ))
                        break
                # note: if the above loop ends without a break, that means that
                # no candidate could be added / yielded by skipping sub-sequence
                # chars

        candidates = new_candidates

    for cand in candidates:
        # note: index + 1 == length(sequence)
        n_skipped = subseq_len - cand.subseq_index
        if cand.n_dels + n_skipped <= max_deletions and \
           cand.l_dist + n_skipped <= max_l_dist:
            yield make_match(cand.start, index + 1, cand.l_dist + n_skipped)


try:
    from fuzzysearch._generic_search import \
        c_find_near_matches_generic_linear_programming as c_fnm_generic_lp
except ImportError:
    find_near_matches_generic_linear_programming = \
        _find_near_matches_generic_linear_programming
else:
    @wraps(_find_near_matches_generic_linear_programming)
    def find_near_matches_generic_linear_programming(subsequence, sequence, search_params):
        try:
            for match in c_fnm_generic_lp(subsequence, sequence, search_params):
                yield match
        except (TypeError, UnicodeEncodeError):
            for match in _find_near_matches_generic_linear_programming(
                    subsequence, sequence, search_params):
                yield match


def find_near_matches_generic_ngrams(subsequence, sequence, search_params):
    """search for near-matches of subsequence in sequence

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the maximum allowed number of character substitutions
    * the maximum allowed number of new characters inserted
    * and the maximum allowed number of character deletions
    * the total number of substitutions, insertions and deletions
    """
    if not subsequence:
        raise ValueError('Given subsequence is empty!')

    max_l_dist = search_params.max_l_dist

    # optimization: prepare some often used things in advance
    subseq_len = len(subsequence)
    seq_len = len(sequence)

    ngram_len = subseq_len // (max_l_dist + 1)
    if ngram_len == 0:
        raise ValueError('the subsequence length must be greater than max_l_dist')

    for ngram_start in range(0, subseq_len - ngram_len + 1, ngram_len):
        ngram_end = ngram_start + ngram_len
        start_index = max(0, ngram_start - max_l_dist)
        end_index = min(seq_len, seq_len - subseq_len + ngram_end + max_l_dist)
        for index in search_exact(subsequence[ngram_start:ngram_end], sequence, start_index, end_index):
            # try to expand left and/or right according to n_ngram
            for match in find_near_matches_generic_linear_programming(
                subsequence, sequence[max(0, index - ngram_start - max_l_dist):index - ngram_start + subseq_len + max_l_dist],
                search_params,
            ):
                yield attr.evolve(match,
                    start=match.start + max(0, index - ngram_start - max_l_dist),
                    end=match.end + max(0, index - ngram_start - max_l_dist),
                )


def has_near_match_generic_ngrams(subsequence, sequence, search_params):
    """search for near-matches of subsequence in sequence

    This searches for near-matches, where the nearly-matching parts of the
    sequence must meet the following limitations (relative to the subsequence):

    * the maximum allowed number of character substitutions
    * the maximum allowed number of new characters inserted
    * and the maximum allowed number of character deletions
    * the total number of substitutions, insertions and deletions
    """
    for match in find_near_matches_generic_ngrams(subsequence, sequence, search_params):
        return True
    return False


class GenericSearch(FuzzySearchBase):
    @classmethod
    def search(cls, subsequence, sequence, search_params):
        for match in find_near_matches_generic(subsequence, sequence,
                                               search_params):
            yield match

    @classmethod
    def consolidate_matches(cls, matches):
        return consolidate_overlapping_matches(matches)

    @classmethod
    def extra_items_for_chunked_search(cls, subsequence, search_params):
        return max(
            x for x in [search_params.max_l_dist,
                        search_params.max_insertions]
            if x is not None
        )



import unittest

from tests.compat import b


class TestMemmemBase(object):
    def search(self, sequence, subsequence):
        raise NotImplementedError

    def test_empty_sequence(self):
        self.assertEqual(self.search('PATTERN', ''), None)

    def test_empty_subsequence(self):
        self.assertEqual(self.search('', 'TEXT'), 0)

    def test_match_identical_sequence(self):
        self.assertEqual(self.search('PATTERN', 'PATTERN'), 0)

    def test_shorter_sequence(self):
        self.assertEqual(self.search('abcd', 'abc'), None)

    def test_substring(self):
        substring = 'PATTERN'
        text = 'aaaaaaaaaaPATTERNaaaaaaaaa'
        self.assertEqual(self.search(substring, text), 10)

    def test_double_first_item(self):
        self.assertEqual(self.search('def', 'abcddefg'), 4)

    def test_missing_second_item(self):
        self.assertEqual(self.search('bde', 'abcdefg'), None)

    def test_completely_different(self):
        self.assertEqual(self.search('abc', 'def'), None)

    def test_startswith(self):
        self.assertEqual(self.search('abc', 'abcd'), 0)

    def test_endswith(self):
        self.assertEqual(self.search('bcd', 'abcd'), 1)

    def test_first_subseq_char_not_in_seq(self):
        self.assertEqual(self.search('xa', 'abcd'), None)

    def test_multiple_appearances(self):
        self.assertEqual(self.search('abc', 'xxxabcxxxabcxxxabcxxx'), 3)

    def test_single_letter_subseq(self):
        self.assertEqual(self.search('a', 'abc'), 0)
        self.assertEqual(self.search('b', 'abc'), 1)
        self.assertEqual(self.search('c', 'abc'), 2)
        self.assertEqual(self.search('a', 'xxx'), None)
        self.assertEqual(self.search('a', 'x'*100 + 'a' + 'x'*100), 100)
        self.assertEqual(self.search('a', ''), None)

    def test_subseq_lengths(self):
        import string
        letters = string.ascii_lowercase + string.ascii_uppercase
        for subseq_length in range(2, 100):
            subseq = letters[:subseq_length]
            self.assertEqual(self.search(subseq, letters), 0)
            self.assertEqual(self.search(subseq, 'a!'*50 + letters), 100)
            self.assertEqual(self.search(subseq, '!'*100), None)
            self.assertEqual(self.search(subseq, '!'), None)
            self.assertEqual(self.search(subseq, '!'*50 + 'a' + '!'*50), None)
            self.assertEqual(self.search(subseq, '!'*50 + 'a'), None)
            self.assertEqual(self.search(subseq, subseq[:-1]), None)
            self.assertEqual(self.search(subseq, subseq[:-1] * 2), None)
            self.assertEqual(self.search(subseq, subseq[:-1] + 'aa'), None)
            self.assertEqual(self.search(subseq, 'aa' + subseq[:-1]), None)
            self.assertEqual(self.search(subseq, 'aa' + subseq[:-1] + 'aa'), None)

try:
    from fuzzysearch._pymemmem import simple_memmem, wordlen_memmem
except ImportError:
    pass
else:
    class TestSimpleMemmem(TestMemmemBase, unittest.TestCase):
        def search(self, subsequence, sequence):
            return simple_memmem(b(subsequence), b(sequence))

    class TestWordlenMemmem(TestMemmemBase, unittest.TestCase):
        def search(self, subsequence, sequence):
            return wordlen_memmem(b(subsequence), b(sequence))
